# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import codecs
import contextlib
import datetime
import json
import logging
import os
from typing import Any, Generator, List, Tuple

from volatility3.framework import constants, exceptions, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import physical
from volatility3.framework.layers import registry as registry_layer
from volatility3.framework.renderers import conversion, format_hints
from volatility3.framework.symbols import intermed
from volatility3.plugins.windows.registry import hivelist
from volatility3.plugins import timeliner

vollog = logging.getLogger(__name__)


class UserAssist(interfaces.plugins.PluginInterface, timeliner.TimeLinerInterface):
    """Print userassist registry keys and information."""

    _required_framework_version = (2, 0, 0)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._userassist_size = 0
        self._userassist_type_name = "_VOL_USERASSIST_TYPES_7"
        self._reg_table_name = None
        self._win7 = None
        # taken from http://msdn.microsoft.com/en-us/library/dd378457%28v=vs.85%29.aspx
        try:
            with open(
                os.path.join(os.path.dirname(__file__), "userassist.json"), "rb"
            ) as fp:
                self._folder_guids = json.load(fp)
        except OSError:
            vollog.error("Usersassist data file not found")

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.IntRequirement(
                name="offset", description="Hive Offset", default=None, optional=True
            ),
            requirements.VersionRequirement(
                name="hivelist", component=hivelist.HiveList, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="timeliner",
                component=timeliner.TimeLinerInterface,
                version=(1, 0, 0),
            ),
        ]

    def parse_userassist_data(self, reg_val):
        """Reads the raw data of a _CM_KEY_VALUE and returns a dict of
        userassist fields."""

        item = {
            "id": renderers.UnparsableValue(),
            "count": renderers.UnparsableValue(),
            "focus": renderers.UnparsableValue(),
            "time": renderers.UnparsableValue(),
            "lastupdated": renderers.UnparsableValue(),
            "rawdata": renderers.UnparsableValue(),
        }

        userassist_data = reg_val.decode_data()

        if userassist_data is None:
            return item

        item["rawdata"] = userassist_data

        if self._win7 is None:
            # if OS is still unknown at this point, return the default item which just has the rawdata
            return item

        if len(userassist_data) < self._userassist_size:
            return item

        userassist_layer_name = self.context.layers.free_layer_name("userassist_buffer")
        buffer = physical.BufferDataLayer(
            self.context, self._config_path, userassist_layer_name, userassist_data
        )
        self.context.add_layer(buffer)
        userassist_obj = self.context.object(
            object_type=self._reg_table_name
            + constants.BANG
            + self._userassist_type_name,
            layer_name=userassist_layer_name,
            offset=0,
        )

        if self._win7:
            item["id"] = renderers.NotApplicableValue()
            item["count"] = int(userassist_obj.Count)

            seconds = (userassist_obj.FocusTime + 500) / 1000.0
            time = (
                datetime.timedelta(seconds=seconds)
                if seconds > 0
                else userassist_obj.FocusTime
            )
            item["focus"] = int(userassist_obj.FocusCount)
            item["time"] = str(time)

        else:
            item["id"] = int(userassist_obj.ID)
            item["count"] = int(
                userassist_obj.CountStartingAtFive
                if userassist_obj.CountStartingAtFive < 5
                else userassist_obj.CountStartingAtFive - 5
            )
            item["focus"] = renderers.NotApplicableValue()
            item["time"] = renderers.NotApplicableValue()

        item["lastupdated"] = conversion.wintime_to_datetime(
            userassist_obj.LastUpdated.QuadPart
        )

        return item

    def _determine_userassist_type(self) -> None:
        """Determine the userassist type and size depending on the OS
        version."""

        if self._win7 is True:
            self._userassist_type_name = "_VOL_USERASSIST_TYPES_7"
        elif self._win7 is False:
            self._userassist_type_name = "_VOL_USERASSIST_TYPES_XP"

        self._userassist_size = self.context.symbol_space.get_type(
            self._reg_table_name + constants.BANG + self._userassist_type_name
        ).size

    def _win7_or_later(self) -> bool:
        # TODO: change this if there is a better way of determining the OS version
        # _KUSER_SHARED_DATA.CookiePad is in Windows 6.1 (Win7) and later
        kernel = self.context.modules[self.config["kernel"]]

        return self.context.symbol_space.get_type(
            kernel.symbol_table_name + constants.BANG + "_KUSER_SHARED_DATA"
        ).has_member("CookiePad")

    def list_userassist(
        self, hive: registry_layer.RegistryHive
    ) -> Generator[Tuple[int, Tuple], None, None]:
        """Generate userassist data for a registry hive."""

        kernel = self.context.modules[self.config["kernel"]]

        hive_name = hive.hive.cast(
            kernel.symbol_table_name + constants.BANG + "_CMHIVE"
        ).get_name()

        if self._win7 is None:
            with contextlib.suppress(exceptions.SymbolError):
                self._win7 = self._win7_or_later()
                # self._win7 will be None and only registry value rawdata will be output

        self._determine_userassist_type()

        try:
            userassist_node_path = hive.get_key(
                "software\\microsoft\\windows\\currentversion\\explorer\\userassist",
                return_list=True,
            )
        except registry_layer.RegistryException as e:
            vollog.warning(
                f"Error accessing UserAssist key in {hive_name} at {hive.hive_offset:#x}: {e}"
            )
            return None
        except KeyError:
            vollog.warning(
                f"UserAssist key not found in {hive_name} at {hive.hive_offset:#x}"
            )
            return None

        if not userassist_node_path:
            vollog.warning("list_userassist did not find a valid node_path (or None)")
            return None

        if not isinstance(userassist_node_path, list):
            vollog.warning("userassist_node_path did not return a list as expected")
            return None
        userassist_node = userassist_node_path[-1]
        # iterate through the GUIDs under the userassist key
        for guidkey in userassist_node.get_subkeys():
            # each guid key should have a Count key in it
            for countkey in guidkey.get_subkeys():
                countkey_path = countkey.get_key_path()
                countkey_last_write_time = conversion.wintime_to_datetime(
                    countkey.LastWriteTime.QuadPart
                )

                # output the parent Count key
                result: Tuple[
                    int,
                    Tuple[
                        format_hints.Hex,
                        Any,
                        Any,
                        Any,
                        Any,
                        Any,
                        Any,
                        Any,
                        Any,
                        Any,
                        Any,
                        Any,
                    ],
                ] = (
                    0,
                    (
                        renderers.format_hints.Hex(hive.hive_offset),
                        hive_name,
                        countkey_path,
                        countkey_last_write_time,
                        "Key",
                        renderers.NotApplicableValue(),
                        renderers.NotApplicableValue(),
                        renderers.NotApplicableValue(),
                        renderers.NotApplicableValue(),
                        renderers.NotApplicableValue(),
                        renderers.NotApplicableValue(),
                        renderers.NotApplicableValue(),
                    ),
                )
                yield result

                # output any subkeys under Count
                for subkey in countkey.get_subkeys():
                    try:
                        subkey_name = subkey.get_name()
                    except (
                        exceptions.InvalidAddressException,
                        registry_layer.RegistryException,
                    ):
                        subkey_name = renderers.UnreadableValue()

                    result = (
                        1,
                        (
                            renderers.format_hints.Hex(hive.hive_offset),
                            hive_name,
                            countkey_path,
                            countkey_last_write_time,
                            "Subkey",
                            subkey_name,
                            renderers.NotApplicableValue(),
                            renderers.NotApplicableValue(),
                            renderers.NotApplicableValue(),
                            renderers.NotApplicableValue(),
                            renderers.NotApplicableValue(),
                            renderers.NotApplicableValue(),
                        ),
                    )
                    yield result

                # output any values under Count
                for value in countkey.get_values():
                    try:
                        value_name = value.get_name()
                    except (
                        exceptions.InvalidAddressException,
                        registry_layer.RegistryException,
                    ):
                        value_name = renderers.UnreadableValue()

                    with contextlib.suppress(UnicodeDecodeError):
                        value_name = codecs.encode(value_name, "rot_13")

                    if self._win7:
                        guid = value_name.split("\\")[0]
                        if guid in self._folder_guids:
                            value_name = value_name.replace(
                                guid, self._folder_guids[guid]
                            )

                    userassist_data_dict = self.parse_userassist_data(value)
                    result = (
                        1,
                        (
                            renderers.format_hints.Hex(hive.hive_offset),
                            hive_name,
                            countkey_path,
                            countkey_last_write_time,
                            "Value",
                            value_name,
                            userassist_data_dict["id"],
                            userassist_data_dict["count"],
                            userassist_data_dict["focus"],
                            userassist_data_dict["time"],
                            userassist_data_dict["lastupdated"],
                            format_hints.HexBytes(userassist_data_dict["rawdata"]),
                        ),
                    )
                    yield result

    def _generator(self):
        hive_offsets = None
        if self.config.get("offset", None) is not None:
            hive_offsets = [self.config.get("offset", None)]

        self._reg_table_name = intermed.IntermediateSymbolTable.create(
            self.context, self._config_path, "windows", "registry"
        )

        # get all the user hive offsets or use the one specified
        for hive in hivelist.HiveList.list_hives(
            context=self.context,
            base_config_path=self.config_path,
            kernel_module_name=self.config["kernel"],
            filter_string="ntuser.dat",
            hive_offsets=hive_offsets,
        ):
            try:
                yield from self.list_userassist(hive)
                continue
            except exceptions.PagedInvalidAddressException as excp:
                vollog.debug(
                    f"Invalid address identified in Hive: {hex(excp.invalid_address)}"
                )
            except exceptions.InvalidAddressException as excp:
                vollog.debug(
                    f"Invalid address identified in lower layer {excp.layer_name}: {excp.invalid_address}"
                )
            except KeyError:
                vollog.debug(
                    "Key '{}' not found in Hive at offset {}.".format(
                        "software\\microsoft\\windows\\currentversion\\explorer\\userassist",
                        hex(hive.hive_offset),
                    )
                )

            # yield UnreadableValues when an exception occurs for a given hive_offset
            result = (
                0,
                (
                    renderers.format_hints.Hex(hive.hive_offset),
                    hive.name if hive.name else renderers.UnreadableValue(),
                    renderers.UnreadableValue(),
                    renderers.UnreadableValue(),
                    renderers.UnreadableValue(),
                    renderers.UnreadableValue(),
                    renderers.UnreadableValue(),
                    renderers.UnreadableValue(),
                    renderers.UnreadableValue(),
                    renderers.UnreadableValue(),
                    renderers.UnreadableValue(),
                    renderers.UnreadableValue(),
                ),
            )
            yield result

    def generate_timeline(self):
        for row in self._generator():
            _depth, row_data = row
            # check the name and the timestamp to not be empty
            if isinstance(row_data[5], str) and not isinstance(
                row_data[10], renderers.NotApplicableValue
            ):
                description = f"UserAssist: {row_data[5]} {row_data[2]} ({row_data[7]})"
                yield (description, timeliner.TimeLinerType.MODIFIED, row_data[10])

    def run(self):
        return renderers.TreeGrid(
            [
                ("Hive Offset", renderers.format_hints.Hex),
                ("Hive Name", str),
                ("Path", str),
                ("Last Write Time", datetime.datetime),
                ("Type", str),
                ("Name", str),
                ("ID", int),
                ("Count", int),
                ("Focus Count", int),
                ("Time Focused", str),
                ("Last Updated", datetime.datetime),
                ("Raw Data", format_hints.HexBytes),
            ],
            self._generator(),
        )



# This file is Copyright 2020 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

from volatility3.framework import renderers, interfaces, exceptions
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility


class VFSevents(interfaces.plugins.PluginInterface):
    """Lists processes that are filtering file system events"""

    _required_framework_version = (2, 0, 0)

    event_types = [
        "CREATE_FILE",
        "DELETE",
        "STAT_CHANGED",
        "RENAME",
        "CONTENT_MODIFIED",
        "EXCHANGE",
        "FINDER_INFO_CHANGED",
        "CREATE_DIR",
        "CHOWN",
        "XATTR_MODIFIED",
        "XATTR_REMOVED",
        "DOCID_CREATED",
        "DOCID_CHANGED",
    ]

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Kernel module for the OS",
                architectures=["Intel32", "Intel64"],
            ),
        ]

    def _generator(self):
        """
        Lists the registered VFS event watching processes
        Also lists which event(s) a process is registered for
        """

        kernel = self.context.modules[self.config["kernel"]]

        watcher_table = kernel.object_from_symbol("watcher_table")

        for watcher in watcher_table:
            if watcher == 0:
                continue

            task_name = utility.array_to_string(watcher.proc_name)
            task_pid = watcher.pid

            events = []

            try:
                event_array = kernel.object(
                    object_type="array",
                    offset=watcher.event_list,
                    absolute=True,
                    count=13,
                    subtype=kernel.get_type("unsigned char"),
                )

            except exceptions.InvalidAddressException:
                continue

            for i, event in enumerate(event_array):
                if event == 1:
                    events.append(self.event_types[i])

            if events != []:
                yield (0, (task_name, task_pid, ",".join(events)))

    def run(self):
        return renderers.TreeGrid(
            [("Name", str), ("PID", int), ("Events", str)], self._generator()
        )



from volatility3.framework.layers import intel

WIN_ARCHS = ["Intel32", "Intel64"]
"""Windows supported architectures"""
WIN_ARCHS_LAYERS = [intel.Intel]
"""Windows supported architectures layers"""

LINUX_ARCHS = ["Intel32", "Intel64"]
"""Linux supported architectures"""
LINUX_ARCHS_LAYERS = [intel.Intel]
"""Linux supported architectures layers"""

MAC_ARCHS = ["Intel32", "Intel64"]
"""Mac supported architectures"""
MAC_ARCHS_LAYERS = [intel.Intel]
"""Mac supported architectures layers"""

FRAMEWORK_ARCHS = ["Intel32", "Intel64"]
"""Framework supported architectures"""
FRAMEWORK_ARCHS_LAYERS = [intel.Intel]
"""Framework supported architectures layers"""



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""The configuration module contains classes and functions for interacting with
the configuration and requirement trees.

Volatility plugins can specify a list of requirements (which may have
subrequirements, thus forming a requirement tree). These requirement
trees can contain values, which are contained in a complementary
configuration tree.  These two trees act as a protocol between the
plugins and users.  The plugins provide requirements that must be
fulfilled, and the users provide configurations values that fulfill
those requirements.  Where the user does not provide sufficient
configuration values, automagic modules may extend the configuration
tree themselves.
"""

import collections.abc
import copy
import json
import logging
import random
import string
import sys
from abc import ABCMeta, abstractmethod
from typing import (
    Any,
    ClassVar,
    Dict,
    Generator,
    Iterator,
    List,
    Optional,
    Type,
    Union,
    Tuple,
    Set,
)

from volatility3 import classproperty, framework
from volatility3.framework import constants, interfaces

CONFIG_SEPARATOR = "."
"""Use to specify the separator between configuration hierarchies"""

vollog = logging.getLogger(__name__)

BasicTypes = (int, bool, bytes, str)
SimpleTypes = Union[int, bool, bytes, str]
ConfigSimpleType = Optional[Union[SimpleTypes, List[SimpleTypes]]]


def path_join(*args) -> str:
    """Joins configuration paths together."""
    # If a path element (particularly the first) is empty, then remove it from the list
    args = tuple(arg for arg in args if arg)
    return CONFIG_SEPARATOR.join(args)


def parent_path(value: str) -> str:
    """Returns the parent configuration path from a configuration path."""
    return CONFIG_SEPARATOR.join(value.split(CONFIG_SEPARATOR)[:-1])


def path_head(value: str) -> str:
    """Return the top of the configuration path"""
    return value.split(CONFIG_SEPARATOR)[-1]


def path_depth(path: str, depth: int = 1) -> str:
    """Returns the `path` up to a certain depth.

    Note that `depth` can be negative (such as `-x`) and will return all
    elements except for the last `x` components
    """
    return path_join(path.split(CONFIG_SEPARATOR)[:depth])


class HierarchicalDict(collections.abc.Mapping):
    """The core of configuration data, it is a mapping class that stores keys
    within itself, and also stores lower hierarchies."""

    def __init__(
        self,
        initial_dict: Optional[Dict[str, "SimpleTypeRequirement"]] = None,
        separator: str = CONFIG_SEPARATOR,
    ) -> None:
        """
        Args:
            initial_dict: A dictionary to populate the HierarchicalDict with initially
            separator: A custom hierarchy separator (defaults to CONFIG_SEPARATOR)
        """
        if not (isinstance(separator, str) and len(separator) == 1):
            raise TypeError(f"Separator must be a one character string: {separator}")
        self._separator = separator
        self._data: Dict[str, ConfigSimpleType] = {}
        self._subdict: Dict[str, HierarchicalDict] = {}
        if isinstance(initial_dict, str):
            initial_dict = json.loads(initial_dict)
        if isinstance(initial_dict, dict):
            for k, v in initial_dict.items():
                self[k] = v
        elif initial_dict is not None:
            raise TypeError(
                f"Initial_dict must be a dictionary or JSON string containing a dictionary: {initial_dict}"
            )

    def __eq__(self, other):
        """Define equality between HierarchicalDicts"""
        return dict(self) == dict(other)

    @property
    def separator(self) -> str:
        """Specifies the hierarchy separator in use in this HierarchyDict."""
        return self._separator

    @property
    def data(self) -> Dict:
        """Returns just the data-containing mappings on this level of the
        Hierarchy."""
        return self._data.copy()

    def _key_head(self, key: str) -> str:
        """Returns the first division of a key based on the dict separator, or
        the full key if the separator is not present."""
        if self.separator in key:
            return key[: key.index(self.separator)]
        else:
            return key

    def _key_tail(self, key: str) -> str:
        """Returns all but the first division of a key based on the dict
        separator, or None if the separator is not in the key."""
        if self.separator in key:
            return key[key.index(self.separator) + 1 :]
        return ""

    def __iter__(self) -> Iterator[Any]:
        """Returns an iterator object that supports the iterator protocol."""
        return self.generator()

    def generator(self) -> Generator[str, None, None]:
        """A generator for the data in this level and lower levels of this
        mapping.

        Returns:
            Returns each item in the top level data, and then all subkeys in a depth first order
        """
        for key in self._data:
            yield key
        for subdict_key in self._subdict:
            for key in self._subdict[subdict_key]:
                yield subdict_key + self.separator + key

    def __getitem__(self, key: str) -> ConfigSimpleType:
        """Gets an item, traversing down the trees to get to the final
        value."""
        try:
            if self.separator in key:
                subdict = self._subdict[self._key_head(key)]
                return subdict[self._key_tail(key)]
            else:
                return self._data[key]
        except KeyError:
            raise KeyError(key)

    def __setitem__(self, key: str, value: Any) -> None:
        """Sets an item or creates a subdict and sets the item within that."""
        self._setitem(key, value)

    def _setitem(self, key: str, value: Any, is_data: bool = True) -> None:
        """Set an item or appends a whole subtree at a key location."""
        if self.separator in key:
            subdict = self._subdict.get(
                self._key_head(key), HierarchicalDict(separator=self.separator)
            )
            subdict._setitem(self._key_tail(key), value, is_data)
            self._subdict[self._key_head(key)] = subdict
        else:
            if is_data:
                self._data[key] = self._sanitize_value(value)
            else:
                if not isinstance(value, HierarchicalDict):
                    raise TypeError(
                        f"HierarchicalDicts can only store HierarchicalDicts within their structure: {type(value)}"
                    )
                self._subdict[key] = value

    def _sanitize_value(self, value: Any) -> ConfigSimpleType:
        """Method to ensure all values are standard values and not volatility
        objects containing contexts."""
        if isinstance(value, bool):
            return bool(value)
        elif isinstance(value, int):
            return int(value)
        elif isinstance(value, str):
            return str(value)
        elif isinstance(value, bytes):
            return bytes(value)
        elif isinstance(value, list):
            new_list = []
            for element in value:
                element_value = self._sanitize_value(element)
                if isinstance(element_value, list):
                    raise TypeError(
                        "Configuration list types cannot contain list types"
                    )
                if element_value is not None:
                    new_list.append(element_value)
            return new_list
        elif value is None:
            return None
        else:
            raise TypeError(f"Invalid type stored in configuration: {type(value)}")

    def __delitem__(self, key: str) -> None:
        """Deletes an item from the hierarchical dict."""
        try:
            if self.separator in key:
                subdict = self._subdict[self._key_head(key)]
                del subdict[self._key_tail(key)]
            else:
                del self._data[self._key_head(key)]
        except KeyError:
            raise KeyError(key)

    def __contains__(self, key: Any) -> bool:
        """Determines whether the key is present in the hierarchy."""
        if self.separator in key:
            try:
                subdict = self._subdict[self._key_head(key)]
                return self._key_tail(key) in subdict
            except KeyError:
                return False
        else:
            return key in self._data

    def __len__(self) -> int:
        """Returns the length of all items."""
        return len(self._data) + sum([len(subdict) for subdict in self._subdict])

    def branch(self, key: str) -> "HierarchicalDict":
        """Returns the HierarchicalDict housed under the key.

        This differs from the data property, in that it is directed by the `key`, and all layers under that key are
        returned, not just those in that level.

        Higher layers are not prefixed with the location of earlier layers, so branching a hierarchy containing `a.b.c.d`
        on `a.b` would return a hierarchy containing `c.d`, not `a.b.c.d`.

        Args:
            key: The location within the hierarchy to return higher layers.

        Returns:
            The HierarchicalDict underneath the specified key (not just the data at that key location in the tree)
        """
        try:
            if self.separator in key:
                return self._subdict[self._key_head(key)].branch(self._key_tail(key))
            else:
                return self._subdict[key]
        except KeyError:
            self._setitem(
                key=key, value=HierarchicalDict(separator=self.separator), is_data=False
            )
        return HierarchicalDict()

    def splice(self, key: str, value: "HierarchicalDict") -> None:
        """Splices an existing HierarchicalDictionary under a specific key.

        This can be thought of as an inverse of :func:`branch`, although
        `branch` does not remove the requested hierarchy, it simply
        returns it.
        """
        if not isinstance(key, str) or not isinstance(value, HierarchicalDict):
            raise TypeError("Splice requires a string key and HierarchicalDict value")
        self._setitem(key, value, False)

    def merge(
        self, key: str, value: "HierarchicalDict", overwrite: bool = False
    ) -> None:
        """Acts similarly to splice, but maintains previous values.

        If overwrite is true, then entries in the new value are used over those that exist within key already

        Args:
            key: The location within the hierarchy at which to merge the `value`
            value: HierarchicalDict to be merged under the key node
            overwrite: A boolean defining whether the value will be overwritten if it already exists
        """
        if not isinstance(key, str) or not isinstance(value, HierarchicalDict):
            raise TypeError("Splice requires a string key and HierarchicalDict value")
        for item in dict(value):
            if self.get(key + self._separator + item, None) is not None:
                if overwrite:
                    self[key + self._separator + item] = value[item]
            else:
                self[key + self._separator + item] = value[item]

    def clone(self) -> "HierarchicalDict":
        """Duplicates the configuration, allowing changes without affecting the
        original.

        Returns:
            A duplicate HierarchicalDict of this object
        """
        return copy.deepcopy(self)

    def __str__(self) -> str:
        """Turns the Hierarchical dict into a string representation."""
        return json.dumps(
            dict([(key, self[key]) for key in sorted(self.generator())]), indent=2
        )


class RequirementInterface(metaclass=ABCMeta):
    """Class that defines a requirement.

    A requirement is a means for plugins and other framework components to request specific configuration data.
    Requirements can either be simple types (such as
    :class:`~volatility3.framework.configuration.requirements.SimpleTypeRequirement`,
    :class:`~volatility3.framework.configuration.requirements.IntRequirement`,
    :class:`~volatility3.framework.configuration.requirements.BytesRequirement` and
    :class:`~volatility3.framework.configuration.requirements.StringRequirement`) or complex types (such
    as :class:`TranslationLayerRequirement`, :class:`SymbolTableRequirement` and :class:`ClassRequirement`
    """

    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        default: ConfigSimpleType = None,
        optional: bool = False,
    ) -> None:
        """

        Args:
            name: The name of the requirement
            description: A short textual description of the requirement
            default: The default value for the requirement if no value is provided
            optional: Whether the requirement must be satisfied or not
        """
        super().__init__()
        if CONFIG_SEPARATOR in name:
            raise ValueError(
                f"Name cannot contain the config-hierarchy divider ({CONFIG_SEPARATOR})"
            )
        self._name = name
        self._description = description or ""
        self._default = default
        self._optional = optional
        self._requirements: Dict[str, RequirementInterface] = {}

    def __repr__(self) -> str:
        return "<" + self.__class__.__name__ + ": " + self.name + ">"

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        for name in self.__dict__:
            if other.__dict__.get(name, None) != self.__dict__[name]:
                return False
        return True

    @property
    def name(self) -> str:
        """The name of the Requirement.

        Names cannot contain CONFIG_SEPARATOR ('.' by default) since
        this is used within the configuration hierarchy.
        """
        return self._name

    @property
    def description(self) -> str:
        """A short description of what the Requirement is designed to affect or
        achieve."""
        return self._description

    @property
    def default(self) -> ConfigSimpleType:
        """Returns the default value if one is set."""
        return self._default

    @property
    def optional(self) -> bool:
        """Whether the Requirement is optional or not."""
        return self._optional

    @optional.setter
    def optional(self, value) -> None:
        """Sets the optional value for a requirement."""
        self._optional = bool(value)

    def config_value(
        self,
        context: "interfaces.context.ContextInterface",
        config_path: str,
        default: ConfigSimpleType = None,
    ) -> ConfigSimpleType:
        """Returns the value for this Requirement from its config path.

        Args:
            context: the configuration store to find the value for this requirement
            config_path: the configuration path of the instance of the requirement to be recovered
            default: a default value to provide if the requirement's configuration value is not found
        """
        return context.config.get(config_path, default)

    # Child operations
    @property
    def requirements(self) -> Dict[str, "RequirementInterface"]:
        """Returns a dictionary of all the child requirements, indexed by
        name."""
        return self._requirements.copy()

    def add_requirement(self, requirement: "RequirementInterface") -> None:
        """Adds a child to the list of requirements.

        Args:
            requirement: The requirement to add as a child-requirement
        """
        self._requirements[requirement.name] = requirement

    def remove_requirement(self, requirement: "RequirementInterface") -> None:
        """Removes a child from the list of requirements.

        Args:
            requirement: The requirement to remove as a child-requirement
        """
        del self._requirements[requirement.name]

    def unsatisfied_children(
        self, context: "interfaces.context.ContextInterface", config_path: str
    ) -> Dict[str, "RequirementInterface"]:
        """Method that will validate all child requirements.

        Args:
            context: the context containing the configuration data for this requirement
            config_path: the configuration path of this instance of the requirement

        Returns:
            A dictionary of full configuration paths for each unsatisfied child-requirement
        """
        result = {}
        for requirement in self.requirements.values():
            if not requirement.optional:
                subresult = requirement.unsatisfied(
                    context, path_join(config_path, self._name)
                )
                result.update(subresult)
        return result

    # Validation routines
    @abstractmethod
    def unsatisfied(
        self, context: "interfaces.context.ContextInterface", config_path: str
    ) -> Dict[str, "RequirementInterface"]:
        """Method to validate the value stored at config_path for the
        configuration object against a context.

        Returns a list containing its own name (or multiple unsatisfied requirement names) when invalid

        Args:
            context: The context object containing the configuration for this requirement
            config_path: The configuration path for this requirement to test satisfaction

        Returns:
            A dictionary of configuration-paths to requirements that could not be satisfied
        """


class SimpleTypeRequirement(RequirementInterface):
    """Class to represent a single simple type (such as a boolean, a string, an
    integer or a series of bytes)"""

    instance_type: ClassVar[Type] = bool

    def add_requirement(self, requirement: RequirementInterface):
        """Always raises a TypeError as instance requirements cannot have
        children."""
        raise TypeError("Instance Requirements cannot have subrequirements")

    def remove_requirement(self, requirement: RequirementInterface):
        """Always raises a TypeError as instance requirements cannot have
        children."""
        raise TypeError("Instance Requirements cannot have subrequirements")

    def unsatisfied(
        self, context: "interfaces.context.ContextInterface", config_path: str
    ) -> Dict[str, RequirementInterface]:
        """Validates the instance requirement based upon its
        `instance_type`."""
        config_path = path_join(config_path, self.name)
        value = self.config_value(context, config_path, self.default)
        if not isinstance(value, self.instance_type):
            vollog.log(
                constants.LOGLEVEL_V,
                f"TypeError - {self.name} requirements only accept {self.instance_type.__name__} type: {repr(value)}",
            )
            return {config_path: self}
        return {}


class ClassRequirement(RequirementInterface):
    """Requires a specific class.

    This is used as means to serialize specific classes for
    :class:`TranslationLayerRequirement` and
    :class:`SymbolTableRequirement` classes.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._cls = None

    def __eq__(self, other):
        # We can just use super because it checks all member of `__dict__`
        # This appeases LGTM and does the right thing
        return super().__eq__(other)

    @property
    def cls(self) -> Optional[Type]:
        """Contains the actual chosen class based on the configuration value's
        class name."""
        return self._cls

    def unsatisfied(
        self, context: "interfaces.context.ContextInterface", config_path: str
    ) -> Dict[str, RequirementInterface]:
        """Checks to see if a class can be recovered."""
        config_path = path_join(config_path, self.name)

        value = self.config_value(context, config_path, self.default)
        self._cls = None
        if value is not None and isinstance(value, str):
            if "." in value:
                # TODO: consider importing the prefix
                module = sys.modules.get(value[: value.rindex(".")], None)
                class_name = value[value.rindex(".") + 1 :]
                if hasattr(module, class_name):
                    self._cls = getattr(module, class_name)
            else:
                if value in globals():
                    self._cls = globals()[value]
        if self._cls is None:
            return {config_path: self}
        return {}


class ConstructableRequirementInterface(RequirementInterface):
    """Defines a Requirement that can be constructed based on their own
    requirements.

    This effectively offers a means for serializing specific python
    types, to be reconstructed based on simple configuration data.  Each
    constructable records a `class` requirement, which indicates the
    object that will be constructed.  That class may have its own
    requirements (which is why validation of a ConstructableRequirement
    must happen after the class configuration value has been provided).
    These values are then provided to the object's constructor by name
    as arguments (as well as the standard `context` and `config_path`
    arguments).
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.add_requirement(
            ClassRequirement("class", "Class of the constructable requirement")
        )
        self._current_class_requirements: Set[Any] = set()

    def __eq__(self, other):
        # We can just use super because it checks all member of `__dict__`
        # This appeases LGTM and does the right thing
        return super().__eq__(other)

    @abstractmethod
    def construct(
        self, context: "interfaces.context.ContextInterface", config_path: str
    ) -> None:
        """Method for constructing within the context any required elements
        from subrequirements.

        Args:
            context: The context object containing the configuration data for the constructable
            config_path: The configuration path for the specific instance of this constructable
        """

    def _validate_class(
        self, context: "interfaces.context.ContextInterface", config_path: str
    ) -> None:
        """Method to check if the class Requirement is valid and if so populate
        the other requirements (but no need to validate, since we're invalid
        already)

        Args:
             context: The context object containing the configuration data for the constructable
             config_path: The configuration path for the specific instance of this constructable
        """
        class_req = self.requirements["class"]
        subreq_config_path = path_join(config_path, self.name)
        if not class_req.unsatisfied(context, subreq_config_path) and isinstance(
            class_req, ClassRequirement
        ):
            # We have a class, and since it's validated we can construct our requirements from it
            if issubclass(class_req.cls, ConfigurableInterface):
                # In case the class has changed, clear out the old requirements
                for old_req in self._current_class_requirements.copy():
                    del self._requirements[old_req]
                    self._current_class_requirements.remove(old_req)
                # And add the new ones
                for requirement in class_req.cls.get_requirements():
                    self._current_class_requirements.add(requirement.name)
                    self.add_requirement(requirement)

    def _construct_class(
        self,
        context: "interfaces.context.ContextInterface",
        config_path: str,
        requirement_dict: Optional[Dict[str, object]] = None,
    ) -> Optional["interfaces.objects.ObjectInterface"]:
        """Constructs the class, handing args and the subrequirements as
        parameters to __init__"""
        if self.requirements["class"].unsatisfied(context, config_path):
            return None

        if not isinstance(self.requirements["class"], ClassRequirement):
            return None
        cls = self.requirements["class"].cls

        if cls is None:
            return None

        # These classes all have a name property
        # We could subclass this out as a NameableInterface, but it seems a little excessive
        # FIXME: We can't test this, because importing the other interfaces causes all kinds of import loops
        # if not issubclass(cls, [interfaces.layers.TranslationLayerInterface,
        #                         interfaces.symbols.SymbolTableInterface]):
        #     return None

        if requirement_dict is None:
            requirement_dict = {}

        # Fulfillment must happen, exceptions happening here mean the requirements aren't correct
        # and these need to be raised and fixed, rather than caught and ignored
        obj = cls(**requirement_dict)
        context.config[config_path] = obj.name
        return obj


class ConfigurableRequirementInterface(RequirementInterface):
    """Simple Abstract class to provide build_required_config."""

    @abstractmethod
    def build_configuration(
        self,
        context: "interfaces.context.ContextInterface",
        config_path: str,
        value: Any,
    ) -> HierarchicalDict:
        """Proxies to a ConfigurableInterface if necessary."""


class ConfigurableInterface(metaclass=ABCMeta):
    """Class to allow objects to have requirements and read configuration data
    from the context config tree."""

    def __init__(
        self, context: "interfaces.context.ContextInterface", config_path: str
    ) -> None:
        """Basic initializer that allows configurables to access their own
        config settings."""
        super().__init__()
        self._context = context
        self._config_path = config_path
        self._config_cache: Optional[HierarchicalDict] = None

    @property
    def context(self) -> "interfaces.context.ContextInterface":
        """The context object that this configurable belongs to/configuration
        is stored in."""
        return self._context

    @property
    def config_path(self) -> str:
        """The configuration path on which this configurable lives."""
        return self._config_path

    @config_path.setter
    def config_path(self, value: str) -> None:
        """The configuration path on which this configurable lives."""
        self._config_path = value
        self._config_cache = None

    @property
    def config(self) -> HierarchicalDict:
        """The Hierarchical configuration Dictionary for this Configurable
        object."""
        if not hasattr(self, "_config_cache") or self._config_cache is None:
            self._config_cache = self._context.config.branch(self._config_path)
        return self._config_cache

    def build_configuration(self) -> HierarchicalDict:
        """Constructs a HierarchicalDictionary of all the options required to
        build this component in the current context.

        Ensures that if the class has been created, it can be recreated
        using the configuration built Inheriting classes must override
        this to ensure any dependent classes update their configurations
        too
        """
        result = HierarchicalDict()
        for req in self.get_requirements():
            value = self.config.get(req.name, None)
            # Do not include the name of constructed classes
            if value is not None and not isinstance(
                req, ConstructableRequirementInterface
            ):
                result[req.name] = value
            if isinstance(req, ConfigurableRequirementInterface):
                if value is not None:
                    result.splice(
                        req.name,
                        req.build_configuration(self.context, self.config_path, value),
                    )
        return result

    @classmethod
    def get_requirements(cls) -> List[RequirementInterface]:
        """Returns a list of RequirementInterface objects required by this
        object."""
        return []

    @classmethod
    def unsatisfied(
        cls, context: "interfaces.context.ContextInterface", config_path: str
    ) -> Dict[str, RequirementInterface]:
        """Returns a list of the names of all unsatisfied requirements.

        Since a satisfied set of requirements will return [], it can be used in tests as follows:

        .. code-block:: python

            unmet = configurable.unsatisfied(context, config_path)
            if unmet:
                raise RuntimeError("Unsatisfied requirements: {}".format(unmet)
        """
        result = {}
        for requirement in cls.get_requirements():
            if not requirement.optional:
                subresult = requirement.unsatisfied(context, config_path)
                result.update(subresult)
        return result

    @classmethod
    def make_subconfig(
        cls,
        context: "interfaces.context.ContextInterface",
        base_config_path: str,
        **kwargs,
    ) -> str:
        """Convenience function to allow constructing a new randomly generated
        sub-configuration path, containing each element from kwargs.

        Args:
            context: The context in which to store the new configuration
            base_config_path: The base configuration path on which to build the new configuration
            kwargs: Keyword arguments that are used to populate the new configuration path

        Returns:
            str: The newly generated full configuration path
        """
        random_config_dict = "".join(
            random.SystemRandom().choices(string.ascii_uppercase + string.digits, k=8)
        )
        new_config_path = path_join(base_config_path, random_config_dict)
        # TODO: Check that the new_config_path is empty, although it's not critical if it's not since the values are merged in

        # This should check that each k corresponds to a requirement and each v is of the appropriate type
        # This would require knowledge of the new configurable itself to verify, and they should do validation in the
        # constructor anyway, however, to prevent bad types getting into the config tree we just verify that v is a basic type
        for k, v in kwargs.items():
            if not isinstance(v, BasicTypes):
                raise TypeError(
                    "Config values passed to make_subconfig can only be simple types"
                )
            context.config[path_join(new_config_path, k)] = v

        return new_config_path


class VersionableInterface:
    """A class that allows version checking so that plugins can request specific versions of components they made need

    This currently includes other Plugins and scanners, but may be extended in the future

    All version number should use semantic versioning
    """

    _version: Tuple[int, int, int] = (0, 0, 0)
    _required_framework_version: Tuple[int, int, int] = (0, 0, 0)

    def __init__(self, *args, **kwargs):
        framework.require_interface_version(*self._required_framework_version)
        super().__init__(*args, **kwargs)

    @classproperty
    def version(cls) -> Tuple[int, int, int]:
        """The version of the current interface (classmethods available on the component).

        It is strongly recommended that Semantic Versioning be used (and the default version verification is defined that way):

            MAJOR version when you make incompatible API changes.
            MINOR version when you add functionality in a backwards compatible manner.
            PATCH version when you make backwards compatible bug fixes.
        """
        return cls._version



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
import re
from typing import List

from volatility3.framework import renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins, configuration
from volatility3.framework.layers import scanners
from volatility3.framework.renderers import format_hints
from volatility3.plugins.windows import pslist

vollog = logging.getLogger(__name__)


class VadRegExScan(plugins.PluginInterface):
    """Scans all virtual memory areas for tasks using RegEx."""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 0)
    MAXSIZE_DEFAULT = 128

    @classmethod
    def get_requirements(cls) -> List[configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
            requirements.VersionRequirement(
                name="regex_scanner",
                component=scanners.RegExScanner,
                version=(1, 0, 0),
            ),
            requirements.StringRequirement(
                name="pattern", description="RegEx pattern", optional=False
            ),
            requirements.IntRequirement(
                name="maxsize",
                description="Maximum size in bytes for displayed context",
                default=cls.MAXSIZE_DEFAULT,
                optional=True,
            ),
        ]

    def _generator(self, regex_pattern, procs):
        regex_pattern = bytes(regex_pattern, "UTF-8")
        vollog.debug(f"RegEx Pattern: {regex_pattern}")

        for proc in procs:
            # attempt to create a process layer for each proc
            proc_layer_name = proc.add_process_layer()
            if not proc_layer_name:
                continue

            # get the proc_layer object from the context
            proc_layer = self.context.layers[proc_layer_name]

            # get process sections for scanning
            sections = []
            for vad in proc.get_vad_root().traverse():
                base = vad.get_start()
                if vad.get_size():
                    sections.append((base, vad.get_size()))

            for offset in proc_layer.scan(
                context=self.context,
                scanner=scanners.RegExScanner(regex_pattern),
                sections=sections,
                progress_callback=self._progress_callback,
            ):
                result_data = proc_layer.read(offset, self.MAXSIZE_DEFAULT, pad=True)

                # reapply the regex in order to extract just the match
                regex_result = re.match(regex_pattern, result_data)

                if regex_result:
                    # the match is within the results_data (e.g. it fits within MAXSIZE_DEFAULT)
                    # extract just the match itself
                    regex_match = regex_result.group(0)
                    text_result = str(regex_match, encoding="UTF-8", errors="replace")
                    bytes_result = regex_match
                else:
                    # the match is not with the results_data (e.g. it doesn't fit within MAXSIZE_DEFAULT)
                    text_result = str(result_data, encoding="UTF-8", errors="replace")
                    bytes_result = result_data

                proc_id = proc.UniqueProcessId
                process_name = proc.ImageFileName.cast(
                    "string",
                    max_length=proc.ImageFileName.vol.count,
                    errors="replace",
                )
                yield (
                    0,
                    (
                        proc_id,
                        process_name,
                        format_hints.Hex(offset),
                        text_result,
                        bytes_result,
                    ),
                )

    def run(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))
        procs = pslist.PsList.list_processes(
            context=self.context,
            kernel_module_name=self.config["kernel"],
            filter_func=filter_func,
        )
        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("Offset", format_hints.Hex),
                ("Text", str),
                ("Hex", bytes),
            ],
            self._generator(self.config.get("pattern"), procs),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import collections
import collections.abc
import datetime
import enum
import logging
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    Optional,
    Tuple,
    TypeVar,
    List,
)

from volatility3.framework import constants, exceptions, interfaces, objects

vollog = logging.getLogger(__name__)

SymbolSpaceReturnType = TypeVar(
    "SymbolSpaceReturnType",
    interfaces.objects.Template,
    interfaces.symbols.SymbolInterface,
    Dict[str, Any],
)


class SymbolType(enum.Enum):
    TYPE = 1
    SYMBOL = 2
    ENUM = 3


class SymbolSpace(interfaces.symbols.SymbolSpaceInterface):
    """Handles an ordered collection of SymbolTables.

    This collection is ordered so that resolution of symbols can proceed
    down through the ranks if a namespace isn't specified.
    """

    def __init__(self) -> None:
        super().__init__()
        self._dict: Dict[str, interfaces.symbols.BaseSymbolTableInterface] = (
            collections.OrderedDict()
        )
        # Permanently cache all resolved symbols
        self._resolved: Dict[str, interfaces.objects.Template] = {}
        self._resolved_symbols: Dict[str, interfaces.objects.Template] = {}

    def clear_symbol_cache(self, table_name: Optional[str] = None) -> None:
        """Clears the symbol cache for the specified table name. If no table
        name is specified, the caches of all symbol tables are cleared."""
        table_list: List[interfaces.symbols.BaseSymbolTableInterface] = []
        if table_name is None:
            table_list = list(self._dict.values())
        else:
            table_list.append(self._dict[table_name])
        for table in table_list:
            table.clear_symbol_cache()

    def free_table_name(self, prefix: str = "layer") -> str:
        """Returns an unused table name to ensure no collision occurs when
        inserting a symbol table."""
        count = 1
        while prefix + str(count) in self:
            count += 1
        return prefix + str(count)

    ### Symbol functions

    def get_symbols_by_type(self, type_name: str) -> Iterable[str]:
        """Returns all symbols based on the type of the symbol."""
        for table in self._dict:
            for symbol_name in self._dict[table].get_symbols_by_type(type_name):
                yield table + constants.BANG + symbol_name

    def get_symbols_by_location(
        self, offset: int, size: int = 0, table_name: Optional[str] = None
    ) -> Iterable[str]:
        """Returns all symbols that exist at a specific relative address."""
        table_list: Iterable[interfaces.symbols.BaseSymbolTableInterface] = (
            self._dict.values()
        )
        if table_name is not None:
            if table_name in self._dict:
                table_list = [self._dict[table_name]]
            else:
                table_list = []
        for table in table_list:
            for symbol_name in table.get_symbols_by_location(offset=offset, size=size):
                yield table.name + constants.BANG + symbol_name

    ### Space functions

    def __len__(self) -> int:
        """Returns the number of tables within the space."""
        return len(self._dict)

    def __getitem__(self, i: str) -> Any:
        """Returns a specific table from the space."""
        return self._dict[i]

    def __iter__(self) -> Iterator[str]:
        """Iterates through all available tables in the symbol space."""
        return iter(self._dict)

    def append(self, value: interfaces.symbols.BaseSymbolTableInterface) -> None:
        """Adds a symbol_list to the end of the space."""
        if not isinstance(value, interfaces.symbols.BaseSymbolTableInterface):
            raise TypeError(value)
        if value.name in self._dict:
            self.remove(value.name)
        self._dict[value.name] = value

    def remove(self, key: str) -> None:
        """Removes a named symbol_list from the space."""
        # Reset the resolved list, since we're removing some symbols
        self._resolved = {}
        del self._dict[key]

    def verify_table_versions(
        self,
        producer: str,
        validator: Callable[[Optional[Tuple], Optional[datetime.datetime]], bool],
        tables: Optional[List[str]] = None,
    ) -> bool:
        """Verifies the producer metadata and version of tables

        Args:
            producer: String name of a table producer to have validation performed
            validator: callable that takes an optional version and an optional datetime that returns False if table is invalid

        Returns:
            False if an invalid table was found or True if no invalid table was found
        """
        if tables is None:
            tables = self._dict.keys()
        for table_name in tables:
            table = self._dict[table_name]
            if not table.producer:
                vollog.debug(
                    f"Symbol table {table_name} could not be validated because no producer metadata was found"
                )
                continue
            if table.producer.name == producer:
                # Run the verification
                if not validator(
                    table.producer.version,
                    table.producer.datetime,
                ):
                    vollog.debug(f"Symbol table {table_name} does not pass validator")
                    return False
            else:
                continue
        return True

    ### Resolution functions

    class UnresolvedTemplate(objects.templates.ReferenceTemplate):
        """Class to highlight when missing symbols are present.

        This class is identical to a reference template, but differentiable by its classname.
        It will output a debug log to indicate when it has been instantiated and with what name.

        This class is designed to be output ONLY as part of the SymbolSpace resolution system.
        Individual SymbolTables that cannot resolve a symbol should still return a SymbolError to
        indicate this failure in resolution.
        """

        def __init__(self, type_name: str, **kwargs) -> None:
            vollog.debug(f"Unresolved reference: {type_name}")
            super().__init__(type_name=type_name, **kwargs)

    def _weak_resolve(
        self, resolve_type: SymbolType, name: str
    ) -> SymbolSpaceReturnType:
        """Takes a symbol name and resolves it with ReferentialTemplates."""
        if resolve_type == SymbolType.TYPE:
            get_function = "get_type"
        elif resolve_type == SymbolType.SYMBOL:
            get_function = "get_symbol"
        elif resolve_type == SymbolType.ENUM:
            get_function = "get_enumeration"
        else:
            raise TypeError("Weak_resolve called without a proper SymbolType")

        name_array = name.split(constants.BANG)
        if len(name_array) == 2:
            table_name = name_array[0]
            component_name = name_array[1]
            try:
                return getattr(self._dict[table_name], get_function)(component_name)
            except KeyError as e:
                raise exceptions.SymbolError(
                    component_name,
                    table_name,
                    f"Type {name} references missing Type/Symbol/Enum: {e}",
                )
        raise exceptions.SymbolError(name, None, f"Malformed name: {name}")

    def _iterative_resolve(self, traverse_list):
        """Iteratively resolves a type, populating linked child
        ReferenceTemplates with their properly resolved counterparts."""
        replacements = set()
        # Whole Symbols that still need traversing
        while traverse_list:
            template_traverse_list, traverse_list = (
                [self._resolved[traverse_list[0]]],
                traverse_list[1:],
            )
            # Traverse a single symbol looking for any ReferenceTemplate objects
            while template_traverse_list:
                traverser, template_traverse_list = (
                    template_traverse_list[0],
                    template_traverse_list[1:],
                )
                for child in traverser.children:
                    if isinstance(child, objects.templates.ReferenceTemplate):
                        # If we haven't seen it before, subresolve it and also add it
                        # to the "symbols that still need traversing" list
                        if child.vol.type_name not in self._resolved:
                            traverse_list.append(child.vol.type_name)
                            try:
                                self._resolved[child.vol.type_name] = (
                                    self._weak_resolve(
                                        SymbolType.TYPE, child.vol.type_name
                                    )
                                )
                            except exceptions.SymbolError:
                                self._resolved[child.vol.type_name] = (
                                    self.UnresolvedTemplate(child.vol.type_name)
                                )
                        # Stash the replacement
                        replacements.add((traverser, child))
                    elif child.children:
                        template_traverse_list.append(child)
        for parent, child in replacements:
            parent.replace_child(child, self._resolved[child.vol.type_name])

    def get_type(self, type_name: str) -> interfaces.objects.Template:
        """Takes a symbol name and resolves it.

        This method ensures that all referenced templates (including
        self-referential templates) are satisfied as ObjectTemplates
        """
        # Traverse down any resolutions
        if type_name not in self._resolved:
            self._resolved[type_name] = self._weak_resolve(SymbolType.TYPE, type_name)  # type: ignore
            self._iterative_resolve([type_name])
        if isinstance(self._resolved[type_name], objects.templates.ReferenceTemplate):
            table_name = None
            index = type_name.find(constants.BANG)
            if index > 0:
                table_name, type_name = type_name[:index], type_name[index + 1 :]
            raise exceptions.SymbolError(
                type_name, table_name, f"Unresolvable symbol requested: {type_name}"
            )
        return self._resolved[type_name]

    def get_symbol(self, symbol_name: str) -> interfaces.symbols.SymbolInterface:
        """Look-up a symbol name across all the contained symbol spaces."""
        retval = self._weak_resolve(SymbolType.SYMBOL, symbol_name)
        if symbol_name not in self._resolved_symbols and retval.type is not None:
            self._resolved_symbols[symbol_name] = self._subresolve(retval.type)
        if not isinstance(retval, interfaces.symbols.SymbolInterface):
            table_name = None
            index = symbol_name.find(constants.BANG)
            if index > 0:
                table_name, symbol_name = symbol_name[:index], symbol_name[index + 1 :]
            raise exceptions.SymbolError(
                symbol_name, table_name, f"Unresolvable Symbol: {symbol_name}"
            )
        return retval

    def _subresolve(
        self, object_template: interfaces.objects.Template
    ) -> interfaces.objects.Template:
        """Ensure an ObjectTemplate doesn't contain any ReferenceTemplates"""
        for child in object_template.children:
            if isinstance(child, objects.templates.ReferenceTemplate):
                new_child = self.get_type(child.vol.type_name)
            else:
                new_child = self._subresolve(child)
            object_template.replace_child(old_child=child, new_child=new_child)
        return object_template

    def get_enumeration(self, enum_name: str) -> interfaces.objects.Template:
        """Look-up a set of enumeration choices from a specific symbol
        table."""
        retval = self._weak_resolve(SymbolType.ENUM, enum_name)
        if not isinstance(retval, interfaces.objects.Template):
            table_name = None
            index = enum_name.find(constants.BANG)
            if index > 0:
                table_name, enum_name = enum_name[:index], enum_name[index + 1 :]
            raise exceptions.SymbolError(
                enum_name, table_name, f"Unresolvable Enumeration: {enum_name}"
            )
        return retval

    def _membership(self, member_type: SymbolType, name: str) -> bool:
        """Test for membership of a component within a table."""

        name_array = name.split(constants.BANG)
        if len(name_array) == 2:
            table_name = name_array[0]
            component_name = name_array[1]
        else:
            return False

        if table_name not in self:
            return False
        table = self[table_name]

        if member_type == SymbolType.TYPE:
            return component_name in table.types
        elif member_type == SymbolType.SYMBOL:
            return component_name in table.symbols
        elif member_type == SymbolType.ENUM:
            return component_name in table.enumerations
        return False

    def has_type(self, name: str) -> bool:
        return self._membership(SymbolType.TYPE, name)

    def has_symbol(self, name: str) -> bool:
        return self._membership(SymbolType.SYMBOL, name)

    def has_enumeration(self, name: str) -> bool:
        return self._membership(SymbolType.ENUM, name)


def symbol_table_is_64bit(
    context: interfaces.context.ContextInterface, symbol_table_name: str
) -> bool:
    """Returns a boolean as to whether a particular symbol table within a
    context is 64-bit or not."""
    return (
        context.symbol_space.get_type(
            symbol_table_name + constants.BANG + "pointer"
        ).size
        == 8
    )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import datetime
import logging
from typing import Iterable, Callable, Optional, Tuple

from volatility3.framework import renderers, interfaces, layers, exceptions
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.windows.extensions import pe
from volatility3.plugins import timeliner
from volatility3.plugins.windows import info
from volatility3.plugins.windows import poolscanner
from volatility3.plugins.windows import pslist

vollog = logging.getLogger(__name__)


class PsScan(interfaces.plugins.PluginInterface, timeliner.TimeLinerInterface):
    """Scans for processes present in a particular windows memory image."""

    _required_framework_version = (2, 3, 1)
    _version = (2, 0, 0)

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="timeliner",
                component=timeliner.TimeLinerInterface,
                version=(1, 0, 0),
            ),
            requirements.VersionRequirement(
                name="info", component=info.Info, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="poolscanner", component=poolscanner.PoolScanner, version=(3, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                element_type=int,
                description="Process ID to include (all other processes are excluded)",
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="dump",
                description="Extract listed processes",
                default=False,
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="physical",
                description="Display physical offset instead of virtual",
                default=False,
                optional=True,
            ),
        ]

    @classmethod
    def physical_offset_from_virtual(cls, context, layer_name, proc):
        """Calculate the physical offset from the virtual offset of a process.

        Args:
            context: The context containing layers and modules information.
            layer_name: The name of the layer containing the process memory.
            proc: The process object for which to calculate the physical offset.

        Returns:
            int: The physical offset of the process.
        Raises:
            TypeError: If the primary layer is not an Intel layer.
        """
        memory = context.layers[layer_name]

        if not isinstance(memory, layers.intel.Intel):
            raise TypeError("Primary layer is not an intel layer")

        (_, _, ph_offset, _, _) = list(
            memory.mapping(offset=proc.vol.offset, length=0)
        )[0]

        return ph_offset

    @classmethod
    def create_offset_filter(
        cls,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        offset: Optional[int] = None,
        physical: bool = True,
        exclude: bool = False,
    ) -> Callable[[interfaces.objects.ObjectInterface], bool]:
        """A factory for producing filter functions that filter based on the physical offset of the process.

        Args:
            offset: A number that is the physical offset to be filtered out
            exclude: Accept only tasks that are not the offset argument

        Returns:
            Filter function to be passed to the list of processes.
        """

        def filter_func(_):
            return False

        if offset:
            if physical:
                if exclude:

                    def filter_func(proc):
                        return (
                            cls.physical_offset_from_virtual(context, layer_name, proc)
                            == offset
                        )

                else:

                    def filter_func(proc):
                        return (
                            cls.physical_offset_from_virtual(context, layer_name, proc)
                            != offset
                        )

            else:
                if exclude:

                    def filter_func(proc):
                        return proc.vol.offset == offset

                else:

                    def filter_func(proc):
                        return proc.vol.offset != offset

        return filter_func

    @classmethod
    def scan_processes(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        filter_func: Callable[[interfaces.objects.ObjectInterface], bool] = lambda _: (
            False
        ),
    ) -> Iterable[interfaces.objects.ObjectInterface]:
        """Scans for processes using the poolscanner module and constraints.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            kernel_module_name: The name of the module for the kernel

        Returns:
            A list of processes found by scanning the `layer_name` layer for process pool signatures
        """

        kernel = context.modules[kernel_module_name]

        constraints = poolscanner.PoolScanner.builtin_constraints(
            kernel.symbol_table_name, [b"Pro\xe3", b"Proc"]
        )

        for result in poolscanner.PoolScanner.generate_pool_scan(
            context, kernel_module_name, constraints
        ):
            _constraint, mem_object, _header = result
            if not filter_func(mem_object):
                yield mem_object

    @classmethod
    def virtual_process_from_physical(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        proc: interfaces.objects.ObjectInterface,
    ) -> Optional[interfaces.objects.ObjectInterface]:
        """Returns a virtual process from a physical addressed one

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            kernel_module_name: The name of the module inside the kernel
            proc: the process object with physical address

        Returns:
            A process object on virtual address layer

        """

        ntkrnlmp = context.modules[kernel_module_name]

        version = cls.get_osversion(context, kernel_module_name)

        tleoffset = ntkrnlmp.get_type("_ETHREAD").relative_child_offset(
            "ThreadListEntry"
        )
        # Start out with the member offset
        offsets = [tleoffset]

        # If (and only if) we're dealing with 64-bit Windows 7 SP1
        # then add the other commonly seen member offset to the list
        bits = context.layers[ntkrnlmp.layer_name].bits_per_register
        if version == (6, 1, 7601) and bits == 64:
            offsets.append(tleoffset + 8)

        # Now we can try to bounce back
        for ofs in offsets:
            ethread = ntkrnlmp.object(
                object_type="_ETHREAD",
                offset=proc.ThreadListHead.Flink - ofs,
                absolute=True,
            )

            # Ask for the thread's process to get an _EPROCESS with a virtual address layer
            virtual_process = ethread.owning_process()
            # Sanity check the bounce.
            # This compares the original offset with the new one (translated from virtual layer)
            (_, _, ph_offset, _, _) = list(
                context.layers[ntkrnlmp.layer_name].mapping(
                    offset=virtual_process.vol.offset, length=0
                )
            )[0]
            if virtual_process and proc.vol.offset == ph_offset:
                return virtual_process
        return None

    @classmethod
    def get_osversion(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
    ) -> Tuple[int, int, int]:
        """Returns the complete OS version (MAJ,MIN,BUILD)

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            kernel_module_name: The name of the module for the kernel
        Returns:
            A tuple with (MAJ,MIN,BUILD)
        """
        kuser = info.Info.get_kuser_structure(context, kernel_module_name)
        nt_major_version = int(kuser.NtMajorVersion)
        nt_minor_version = int(kuser.NtMinorVersion)
        vers = info.Info.get_version_structure(context, kernel_module_name)
        build = vers.MinorVersion
        return (nt_major_version, nt_minor_version, build)

    def _generator(self):
        kernel = self.context.modules[self.config["kernel"]]
        pe_table_name = intermed.IntermediateSymbolTable.create(
            self.context, self.config_path, "windows", "pe", class_types=pe.class_types
        )
        memory = self.context.layers[kernel.layer_name]
        if not isinstance(memory, layers.intel.Intel):
            raise TypeError("Primary layer is not an intel layer")

        for proc in self.scan_processes(
            self.context,
            self.config["kernel"],
            filter_func=pslist.PsList.create_pid_filter(self.config.get("pid", None)),
        ):
            file_output = "Disabled"
            if self.config["dump"]:
                # windows 10 objects (maybe others in the future) are already in virtual memory
                if proc.vol.layer_name == kernel.layer_name:
                    vproc = proc
                else:
                    try:
                        vproc = self.virtual_process_from_physical(
                            self.context,
                            self.config["kernel"],
                            proc,
                        )
                    except exceptions.PagedInvalidAddressException:
                        vproc = None

                file_output = "Error outputting file"
                if vproc:
                    file_handle = pslist.PsList.process_dump(
                        self.context,
                        kernel.symbol_table_name,
                        pe_table_name,
                        vproc,
                        self.open,
                    )

                    if file_handle:
                        file_output = file_handle.preferred_filename

            if not self.config["physical"]:
                offset = proc.vol.offset
            else:
                (_, _, offset, _, _) = list(
                    memory.mapping(offset=proc.vol.offset, length=0)
                )[0]

            try:
                yield (
                    0,
                    (
                        proc.UniqueProcessId,
                        proc.InheritedFromUniqueProcessId,
                        proc.ImageFileName.cast(
                            "string",
                            max_length=proc.ImageFileName.vol.count,
                            errors="replace",
                        ),
                        format_hints.Hex(offset),
                        proc.ActiveThreads,
                        proc.get_handle_count(),
                        proc.get_session_id(),
                        proc.get_is_wow64(),
                        proc.get_create_time(),
                        proc.get_exit_time(),
                        file_output,
                    ),
                )
            except exceptions.InvalidAddressException:
                vollog.info(
                    f"Invalid process found at address: {proc.vol.offset:x}. Skipping"
                )

    def generate_timeline(self):
        for row in self._generator():
            _depth, row_data = row
            description = f"Process: {row_data[0]} {row_data[2]} ({row_data[3]})"
            yield (description, timeliner.TimeLinerType.CREATED, row_data[8])
            yield (description, timeliner.TimeLinerType.MODIFIED, row_data[9])

    def run(self):
        offsettype = "(V)" if not self.config["physical"] else "(P)"
        return renderers.TreeGrid(
            [
                ("PID", int),
                ("PPID", int),
                ("ImageFileName", str),
                (f"Offset{offsettype}", format_hints.Hex),
                ("Threads", int),
                ("Handles", int),
                ("SessionId", int),
                ("Wow64", bool),
                ("CreateTime", datetime.datetime),
                ("ExitTime", datetime.datetime),
                ("File output", str),
            ],
            self._generator(),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""Volatility 3 framework."""

# Check the python version to ensure it's suitable
import glob
import sys
import zipfile
import importlib
import inspect
import logging
import os
import traceback
from typing import Any, Dict, Generator, List, Optional, Tuple, Type, TypeVar

from volatility3.framework import constants, interfaces, versionutils

if (
    sys.version_info.major != constants.REQUIRED_PYTHON_VERSION[0]
    or sys.version_info.minor < constants.REQUIRED_PYTHON_VERSION[1]
    or (
        sys.version_info.minor == constants.REQUIRED_PYTHON_VERSION[1]
        and sys.version_info.micro < constants.REQUIRED_PYTHON_VERSION[2]
    )
):
    raise RuntimeError(
        f"Volatility framework requires python version {'.'.join(str(x) for x in constants.REQUIRED_PYTHON_VERSION)} or greater"
    )

# ##
#
# SemVer version scheme
#
# Increment the:
#
#     MAJOR version when you make incompatible API changes,
#     MINOR version when you add functionality in a backwards compatible manner, and
#     PATCH version when you make backwards compatible bug fixes.


def interface_version() -> Tuple[int, int, int]:
    """Provides the so version number of the library."""
    return constants.VERSION_MAJOR, constants.VERSION_MINOR, constants.VERSION_PATCH


vollog = logging.getLogger(__name__)


def require_interface_version(*args) -> None:
    """Checks the required version of a plugin."""
    if not versionutils.matches_required(args, interface_version()):
        raise RuntimeError(
            "Framework interface version {} is incompatible with required version {}".format(
                ".".join(str(x) for x in interface_version()[0:2]),
                ".".join(str(x) for x in args[0:2]),
            )
        )


class NonInheritable:
    def __init__(self, value: Any, cls: Type) -> None:
        self.default_value = value
        self.cls = cls

    def __get__(self, obj: Any, get_type: Type = Optional[None]) -> Any:
        if type is self.cls:
            if hasattr(self.default_value, "__get__"):
                return self.default_value.__get__(obj, get_type)
            return self.default_value
        raise AttributeError


def hide_from_subclasses(cls: Type) -> Type:
    cls.hidden = NonInheritable(True, cls)
    return cls


T = TypeVar("T")


def class_subclasses(cls: Type[T]) -> Generator[Type[T], None, None]:
    """Returns all the (recursive) subclasses of a given class."""
    if not inspect.isclass(cls):
        raise TypeError(f"class_subclasses parameter not a valid class: {cls}")
    for clazz in cls.__subclasses__():
        # The typing system is not clever enough to realize that clazz has a hidden attr after the hasattr check
        if not hasattr(clazz, "hidden") or not clazz.hidden:  # type: ignore
            yield clazz
        yield from class_subclasses(clazz)


def import_files(base_module, ignore_errors: bool = False) -> List[str]:
    """Imports all plugins present under plugins module namespace."""
    failures = []
    if not isinstance(base_module.__path__, list):
        raise TypeError("[base_module].__path__ must be a list of paths")
    vollog.log(
        constants.LOGLEVEL_VVVV,
        f"Importing from the following paths: {', '.join(base_module.__path__)}",
    )
    for path in base_module.__path__:
        for root, _, files in os.walk(path, followlinks=True):
            # TODO: Figure out how to import pycache files
            if root.endswith("__pycache__"):
                continue
            for filename in files:
                if zipfile.is_zipfile(os.path.join(root, filename)):
                    # Use the root to add this to the module path, and sub-traverse the files
                    new_module = base_module
                    premodules = root[len(path) + len(os.path.sep) :].replace(
                        os.path.sep, "."
                    )
                    for component in premodules.split("."):
                        if component:
                            try:
                                new_module = getattr(new_module, component)
                            except AttributeError:
                                failures += [new_module + "." + component]
                    new_module.__path__ = [
                        os.path.join(root, filename)
                    ] + new_module.__path__
                    for ziproot, zipfiles in _zipwalk(os.path.join(root, filename)):
                        for zfile in zipfiles:
                            if _filter_files(zfile):
                                submodule = zfile[: zfile.rfind(".")].replace(
                                    os.path.sep, "."
                                )
                                failures += import_file(
                                    new_module.__name__ + "." + submodule,
                                    os.path.join(path, ziproot, zfile),
                                )
                else:
                    if _filter_files(filename):
                        modpath = os.path.join(
                            root[len(path) + len(os.path.sep) :],
                            filename[: filename.rfind(".")],
                        )
                        submodule = modpath.replace(os.path.sep, ".")
                        failures += import_file(
                            base_module.__name__ + "." + submodule,
                            os.path.join(root, filename),
                            ignore_errors,
                        )

    return failures


def _filter_files(filename: str):
    """Ensures that a filename traversed is an importable python file"""
    return (filename.endswith((".py", ".pyc"))) and not filename.startswith("__")


def import_file(module: str, path: str, ignore_errors: bool = False) -> List[str]:
    """Imports a python file based on an existing module, a submodule and a filepath for error messages

    Args
        module: Module name to be imported
        path: File to be imported from (used for error messages)

    Returns
        List of modules that may have failed to import

    """
    failures = []
    if module not in sys.modules:
        try:
            importlib.import_module(module)
        except ImportError as e:
            vollog.debug(
                "".join(
                    traceback.TracebackException.from_exception(e).format(chain=True)
                )
            )
            vollog.debug(f"Failed to import module {module} based on file: {path}")
            failures.append(module)
            if not ignore_errors:
                raise
    return failures


def _zipwalk(path: str):
    """Walks the contents of a zipfile just like os.walk"""
    zip_results = {}
    with zipfile.ZipFile(path) as archive:
        for file in archive.filelist:
            if not file.is_dir():
                dirlist = zip_results.get(os.path.dirname(file.filename), [])
                dirlist.append(os.path.basename(file.filename))
                zip_results[os.path.join(path, os.path.dirname(file.filename))] = (
                    dirlist
                )
    yield from zip_results.items()


def list_plugins() -> Dict[str, Type[interfaces.plugins.PluginInterface]]:
    plugin_list = {}
    for plugin in class_subclasses(interfaces.plugins.PluginInterface):
        plugin_name = plugin.__module__ + "." + plugin.__name__
        if plugin_name.startswith("volatility3.plugins."):
            plugin_name = plugin_name[len("volatility3.plugins.") :]
        plugin_list[plugin_name] = plugin
    return plugin_list


def clear_cache(complete=True):
    try:
        if complete:
            glob_pattern = "*.cache"
            for cache_filename in glob.glob(
                os.path.join(constants.CACHE_PATH, glob_pattern)
            ):
                os.unlink(cache_filename)
        os.unlink(os.path.join(constants.CACHE_PATH, constants.IDENTIFIERS_FILENAME))
    except FileNotFoundError:
        vollog.log(constants.LOGLEVEL_VVVV, "Attempting to clear a non-existent cache")



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging

from volatility3.framework import deprecation, interfaces
from volatility3.plugins.windows.registry import scheduled_tasks

vollog = logging.getLogger(__name__)


class ScheduledTasks(
    interfaces.plugins.PluginInterface,
    deprecation.PluginRenameClass,
    replacement_class=scheduled_tasks.ScheduledTasks,
    removal_date="2026-09-25",
):
    """Decodes scheduled task information from the Windows registry, including
    information about triggers, actions, run times, and creation times (deprecated)."""

    _required_framework_version = (2, 11, 0)
    _version = (2, 0, 0)



# This file is Copyright 2023 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import enum
import logging
import os
import struct
from typing import Dict, List

from volatility3.framework import constants, exceptions, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import configuration, plugins
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import intermed

vollog = logging.getLogger(__name__)


class VMCSTest(enum.IntFlag):
    VMCS_ABORT_INVALID = enum.auto()
    VMCS_LINK_PTR_IS_NOT_FS = enum.auto()
    VMCS_HOST_CR4_NO_VTX = enum.auto()
    VMCS_CR3_IS_ZERO = enum.auto()
    VMCS_GUEST_CR4_RESERVED = enum.auto()


class PageStartScanner(interfaces.layers.ScannerInterface):
    _version = (1, 0, 0)

    def __init__(self, signatures: List[bytes], page_size: int = 0x1000):
        super().__init__()
        if not len(signatures):
            raise ValueError("No signatures passed to constructor")
        self._siglen = len(signatures[0])
        for item in signatures:
            if len(item) != self._siglen:
                raise ValueError(
                    "Signatures of different lengths passed to PageStartScanner"
                )
        self._signatures = signatures
        self._page_size = page_size

    def __call__(self, data: bytes, data_offset: int):
        """Scans only the start of every page, to see whether a signature is present or not"""
        for page_start in range(
            data_offset % self._page_size, len(data), self._page_size
        ):
            if data[page_start : page_start + self._siglen] in self._signatures:
                yield (
                    page_start + data_offset,
                    data[page_start : page_start + self._siglen],
                )


class Vmscan(plugins.PluginInterface):
    """Scans for Intel VT-d structures and generates VM volatility configs for them"""

    _required_framework_version = (2, 2, 0)
    _version = (1, 0, 0)

    STRICTLY_REQUIRED_TESTS = {
        VMCSTest.VMCS_ABORT_INVALID,
        VMCSTest.VMCS_LINK_PTR_IS_NOT_FS,
        VMCSTest.VMCS_HOST_CR4_NO_VTX,
    }

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.TranslationLayerRequirement(
                name="primary", description="Physical base memory layer"
            ),
            requirements.VersionRequirement(
                name="page_start_scanner",
                component=PageStartScanner,
                version=(1, 0, 0),
            ),
            requirements.IntRequirement(
                name="log-threshold",
                description="Number of criteria failed to log to debug output",
                default=2,
                optional=True,
            ),
        ]

    # Scan for VMCS structures based on the known VMCS structures
    # found in symbols/vmcs directory

    def _gather_vmcs_structures(
        self, context: interfaces.context.ContextInterface, config_path: str
    ) -> Dict[bytes, str]:
        """Enumerate all JSON files containing VMCS information and return the structures
        Signatures can be generated using data extracted using the vmcs_layout tool at
        https://github.com/google/rekall/tree/master/tools/linux/vmcs_layout

        Args:
            context: The volatility context to work against
            config_path: The location to store symbol table configurations under

        Returns:
            A dictionary of pattern bytes to the string representation of the architecture
        """
        filenames = intermed.IntermediateSymbolTable.file_symbol_url(
            os.path.join("generic", "vmcs")
        )
        table_names = []
        for filename in filenames:
            base_name = os.path.basename(filename).split(".")[0]
            table_name = intermed.IntermediateSymbolTable.create(
                context,
                configuration.path_join(config_path, "vmcs"),
                os.path.join("generic", "vmcs"),
                filename=base_name,
            )
            table_names.append(table_name)

        result = {}
        for table_name in table_names:
            symbol_table = context.symbol_space[table_name]
            revision_id = struct.pack(
                "<I", int(symbol_table.get_symbol("revision_id").constant_data)
            )
            result[revision_id] = table_name

        return result

    @classmethod
    def _verify_vmcs_page(
        cls,
        context: interfaces.context.ContextInterface,
        vmcs: interfaces.objects.ObjectInterface,
    ) -> List[str]:
        """Runs tests to verify whether a block of data is a VMCS page
        Some tests based on the Hypervisor Memory Forensics paper by
        Mariano Graziano, Andrea Lanzi and Davide Balzarotti

        Args:
            context: The volatility context to be used for this call
            vmcs: The instantiated VMCS object to verify

        Returns:
            The list of failed criteria that the VMCS did not meet
        """

        # The VMCS should have been constructed on the physical layer (even a nested VMCS)
        physical_layer_name = vmcs.vol.layer_name

        failed_tests: VMCSTest = VMCSTest(0)
        # The abort field must be valid (generally 0, although other abort codes may exist)
        if context.layers[physical_layer_name].read(vmcs.vol.offset + 4, 4) not in [
            b"\x00\x00\x00\x00"
        ]:
            failed_tests |= VMCSTest.VMCS_ABORT_INVALID
        # The vmcs link pointer is supposed to always be set
        if vmcs.vmcs_link_ptr != 0xFFFFFFFFFFFFFFFF:
            failed_tests |= VMCSTest.VMCS_LINK_PTR_IS_NOT_FS
        # To have a VMCS the host needs the VTx bit set in CR4, this can false positive often when all bits are set
        if (vmcs.host_cr4 & 1 << 13) == 0:
            failed_tests |= VMCSTest.VMCS_HOST_CR4_NO_VTX
        # The guest CR3 is *exceptionally* unlikely to be 0 and the guest cr4 is likely to have some bits unset
        if (vmcs.guest_cr3 == 0) or (vmcs.host_cr3 == 0):
            failed_tests |= VMCSTest.VMCS_CR3_IS_ZERO
        # CR4 registers have certain bits reserved that should not be set
        if vmcs.guest_cr4 & 0xFFFFFFFFFF889000:
            failed_tests |= VMCSTest.VMCS_GUEST_CR4_RESERVED

        if failed_tests and failed_tests.name:
            failed_list = failed_tests.name.split("|")
            return failed_list

        return []

    def _generator(self):
        # Gather VMCS structures
        structures = self._gather_vmcs_structures(self.context, self.config_path)
        # Scan memory for them
        layer = self.context.layers[self.config["primary"]]

        # Try to move down to the highest physical layer
        if layer.config.get("memory_layer"):
            layer = self.context.layers[layer.config["memory_layer"]]

        # Run the scan
        for offset, match in layer.scan(
            self.context,
            PageStartScanner(list(structures.keys())),
            self._progress_callback,
        ):
            try:
                vmcs = self.context.object(
                    structures[match] + constants.BANG + "_VMCS",
                    layer.name,
                    offset=offset,
                )
                failed_list = self._verify_vmcs_page(self.context, vmcs)
                if not failed_list:
                    yield (
                        0,
                        (
                            structures[match],
                            format_hints.Hex(vmcs.vol.offset),
                            format_hints.Hex(vmcs.ept),
                            format_hints.Hex(vmcs.guest_cr3),
                        ),
                    )
                    if len(failed_list) <= self.config["log-threshold"]:
                        vollog.debug(
                            f"Potential {structures[match]} VMCS found at {vmcs.vol.offset:x} with failed criteria: {failed_list}"
                        )
            except (exceptions.InvalidAddressException, AttributeError):
                # Not what we're looking for
                continue

    def run(self):
        return renderers.TreeGrid(
            [
                ("Architecture", str),
                ("VMCS Physical offset", format_hints.Hex),
                ("EPT", format_hints.Hex),
                ("Guest CR3", format_hints.Hex),
            ],
            self._generator(),
        )



# This file is Copyright 2020 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0

import hashlib
import json
import logging
import os
import struct
from typing import List

from volatility3.framework import renderers, interfaces, constants, exceptions
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import registry
from volatility3.plugins.windows.registry import hivelist

vollog = logging.getLogger(__name__)


def createservicesid(svc) -> str:
    """Calculate the Service SID"""
    uni = "".join([c + "\x00" for c in svc])
    sha = hashlib.sha1(uni.upper().encode("utf-8")).digest()  # pylint: disable-msg=E1101
    dec = list()
    for i in range(5):
        ## The use of struct here is OK. It doesn't make much sense
        ## to leverage obj.Object inside this loop.
        dec.append(struct.unpack("<I", sha[i * 4 : i * 4 + 4])[0])
    return "S-1-5-80-" + "-".join(str(n) for n in dec)


class GetServiceSIDs(interfaces.plugins.PluginInterface):
    """Lists process token sids."""

    _version = (1, 0, 0)
    _required_framework_version = (2, 0, 0)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Find the sids json path (or raise error if its not in the plugin directory).
        for plugin_dir in constants.PLUGINS_PATH:
            sids_json_file_name = os.path.join(
                plugin_dir, os.path.join("windows", "sids_and_privileges.json")
            )
            if os.path.exists(sids_json_file_name):
                break
        else:
            vollog.log(
                constants.LOGLEVEL_VVV,
                "sids_and_privileges.json file is missing plugin error",
            )
            raise RuntimeError(
                "The sids_and_privileges.json file missed from you plugin directory"
            )

        # Get service sids dictionary (we need only the service sids).
        with open(sids_json_file_name) as file_handle:
            self.servicesids = json.load(file_handle)["service sids"]

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="hivelist", component=hivelist.HiveList, version=(2, 0, 0)
            ),
        ]

    def _generator(self):
        # Get the system hive
        for hive in hivelist.HiveList.list_hives(
            context=self.context,
            base_config_path=self.config_path,
            kernel_module_name=self.config["kernel"],
            filter_string="machine\\system",
            hive_offsets=None,
        ):
            # Get ControlSet\Services.
            try:
                services = hive.get_key(r"CurrentControlSet\Services")
            except (
                KeyError,
                exceptions.InvalidAddressException,
                registry.RegistryException,
            ):
                try:
                    services = hive.get_key(r"ControlSet001\Services")
                except (
                    KeyError,
                    exceptions.InvalidAddressException,
                    registry.RegistryException,
                ):
                    continue

            if services:
                for s in services.get_subkeys():
                    try:
                        sid_name = s.get_name()
                    except (
                        exceptions.InvalidAddressException,
                        registry.RegistryException,
                    ):
                        continue

                    if sid_name not in self.servicesids.values():
                        sid = createservicesid(sid_name)
                        yield (0, (sid, sid_name))

    def run(self):
        return renderers.TreeGrid([("SID", str), ("Service", str)], self._generator())



# This file is Copyright 2023 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import Iterable, List, Tuple

from volatility3.framework import interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.plugins import yarascan
from volatility3.plugins.linux import pslist

vollog = logging.getLogger(__name__)


class VmaYaraScan(interfaces.plugins.PluginInterface):
    """Scans all virtual memory areas for tasks using yara."""

    _required_framework_version = (2, 22, 0)
    _version = (1, 0, 4)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # create a list of requirements for vmayarascan
        vmayarascan_requirements = [
            requirements.ListRequirement(
                name="pid",
                element_type=int,
                description="Process IDs to include (all other processes are excluded)",
                optional=True,
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
            requirements.VersionRequirement(
                name="yarascan", component=yarascan.YaraScan, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="yarascanner", component=yarascan.YaraScanner, version=(2, 0, 0)
            ),
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
        ]

        # get base yarascan requirements for command line options
        yarascan_requirements = yarascan.YaraScan.get_yarascan_option_requirements()

        # return the combined requirements
        return yarascan_requirements + vmayarascan_requirements

    def _generator(self):
        # use yarascan to parse the yara options provided and create the rules
        rules = yarascan.YaraScan.process_yara_options(dict(self.config))

        sanity_check = 1024 * 1024 * 1024  # 1 GB

        # filter based on the pid option if provided
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))
        for task in pslist.PsList.list_tasks(
            context=self.context,
            vmlinux_module_name=self.config["kernel"],
            filter_func=filter_func,
        ):
            # attempt to create a process layer for each task and skip those
            # that cannot (e.g. kernel threads)
            proc_layer_name = task.add_process_layer()
            if not proc_layer_name:
                continue

            # get the proc_layer object from the context
            proc_layer = self.context.layers[proc_layer_name]

            max_vma_size = 0
            vma_maps_to_scan = []
            for start, size in self.get_vma_maps(task):
                if size > sanity_check:
                    vollog.debug(
                        f"VMA at 0x{start:x} over sanity-check size, not scanning"
                    )
                    continue
                max_vma_size = max(max_vma_size, size)
                vma_maps_to_scan.append((start, size))

            if not vma_maps_to_scan:
                vollog.warning(f"No VMAs were found for task {task.tgid}, not scanning")
                continue

            scanner = yarascan.YaraScanner(rules=rules)
            scanner.chunk_size = max_vma_size

            # scan the VMA data (in one contiguous block) with the yarascanner
            for start, size in vma_maps_to_scan:
                for offset, rule_name, name, value in scanner(
                    proc_layer.read(start, size, pad=True), start
                ):
                    layer_data = renderers.LayerData(
                        context=self.context,
                        offset=offset,
                        layer_name=proc_layer.name,
                        length=len(value),
                    )
                    yield (
                        0,
                        (
                            format_hints.Hex(offset),
                            task.tgid,
                            rule_name,
                            name,
                            layer_data,
                        ),
                    )

    @classmethod
    def get_vma_maps(
        cls,
        task: interfaces.objects.ObjectInterface,
    ) -> Iterable[Tuple[int, int]]:
        """Creates a map of start/end addresses for each virtual memory area in a task.

        Args:
            task: The task object of which to read the vmas from

        Returns:
            An iterable of tuples containing start and end addresses for each descriptor
        """
        if task.mm:
            for vma in task.mm.get_vma_iter():
                vm_size = vma.vm_end - vma.vm_start
                yield (vma.vm_start, vm_size)

    def run(self):
        return renderers.TreeGrid(
            [
                ("Offset", format_hints.Hex),
                ("PID", int),
                ("Rule", str),
                ("Component", str),
                ("Value", renderers.LayerData),
            ],
            self._generator(),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""A CommandLine User Interface for the volatility framework.

User interfaces make use of the framework to:
 * determine available plugins
 * request necessary information for those plugins from the user
 * determine what "automagic" modules will be used to populate information the user does not provide
 * run the plugin
 * display the results
"""

import argparse
import inspect
import io
import json
import logging
import os
import sys
import tempfile
import traceback
from typing import Any, Dict, List, Optional, Tuple, Type, Union
from urllib import parse, request

try:
    import argcomplete

    HAS_ARGCOMPLETE = True
except ImportError:
    HAS_ARGCOMPLETE = False

from volatility3.cli import text_filter
import volatility3.plugins
import volatility3.symbols
from volatility3 import framework
from volatility3.cli import text_renderer, volargparse
from volatility3.framework import (
    automagic,
    configuration,
    constants,
    contexts,
    exceptions,
    interfaces,
    plugins,
)
from volatility3.framework.automagic import stacker
from volatility3.framework.configuration import requirements

# Make sure we log everything

rootlog = logging.getLogger()
vollog = logging.getLogger(__name__)
console = logging.StreamHandler()
console.setLevel(logging.WARNING)
formatter = logging.Formatter("%(levelname)-8s %(name)-12s: %(message)s")
# Trim the console down by default
console.setFormatter(formatter)


class PrintedProgress:
    """A progress handler that prints the progress value and the description
    onto the command line."""

    def __init__(self):
        self._max_message_len = 0

    def __call__(self, progress: Union[int, float], description: Optional[str] = None):
        """A simple function for providing text-based feedback.

        .. warning:: Only for development use.

        Args:
            progress: Percentage of progress of the current procedure
        """
        message = f"\rProgress: {round(progress, 2): 7.2f}\t\t{description or ''}"
        message_len = len(message)
        self._max_message_len = max([self._max_message_len, message_len])
        sys.stderr.write(message + (" " * (self._max_message_len - message_len)) + "\r")


class MuteProgress(PrintedProgress):
    """A dummy progress handler that produces no output when called."""

    def __call__(self, progress: Union[int, float], description: Optional[str] = None):
        pass


class CommandLine:
    """Constructs a command-line interface object for users to run plugins."""

    CLI_NAME = os.path.basename(sys.argv[0])  # vol or volatility

    def __init__(self):
        self.setup_logging()
        self.output_dir = None

    @classmethod
    def setup_logging(cls):
        # Delay the setting of vollog for those that want to import volatility3.cli (issue #241)
        rootlog.setLevel(1)
        rootlog.addHandler(console)

    def run(self):
        """Executes the command line module, taking the system arguments,
        determining the plugin to run and then running it."""

        volatility3.framework.require_interface_version(2, 0, 0)

        # Load up system defaults
        delayed_logs, default_config = self.load_system_defaults("vol.json")

        parser = volargparse.HelpfulArgParser(
            add_help=False,
            prog=self.CLI_NAME,
            description="An open-source memory forensics framework",
        )
        parser.add_argument(
            "-h",
            "--help",
            action="help",
            default=argparse.SUPPRESS,
            help=f"Show this help message and exit, for specific plugin options use '{parser.prog} <pluginname> --help'",
        )
        parser.add_argument(
            "-c",
            "--config",
            help="Load the configuration from a json file",
            default=None,
            type=str,
        )
        parser.add_argument(
            "--parallelism",
            help="Enables parallelism (defaults to off if no argument given)",
            nargs="?",
            choices=["processes", "threads", "off"],
            const="processes",
            default=None,
            type=str,
        )
        parser.add_argument(
            "-e",
            "--extend",
            help="Extend the configuration with a new (or changed) setting",
            default=None,
            action="append",
        )
        parser.add_argument(
            "-p",
            "--plugin-dirs",
            help="Semi-colon separated list of paths to find plugins",
            default="",
            type=str,
        )
        parser.add_argument(
            "-s",
            "--symbol-dirs",
            help="Semi-colon separated list of paths to find symbols",
            default="",
            type=str,
        )
        parser.add_argument(
            "-v",
            "--verbosity",
            help="Increase output verbosity",
            default=0,
            action="count",
        )
        parser.add_argument(
            "-l",
            "--log",
            help="Log output to a file as well as the console",
            default=None,
            type=str,
        )
        parser.add_argument(
            "-o",
            "--output-dir",
            help="Directory in which to output any generated files",
            default=os.getcwd(),
            type=str,
        )
        parser.add_argument(
            "-q",
            "--quiet",
            help="Remove progress feedback",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "-f",
            "--file",
            metavar="FILE",
            default=None,
            type=str,
            help="Shorthand for --single-location=file:// if single-location is not defined",
        )
        parser.add_argument(
            "--write-config",
            help="Write configuration JSON file out to config.json",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--save-config",
            help="Save configuration JSON file to a file",
            default=None,
            type=str,
        )
        parser.add_argument(
            "--clear-cache",
            help="Clears out all short-term cached items",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--cache-path",
            help=f"Change the default path ({constants.CACHE_PATH}) used to store the cache",
            default=constants.CACHE_PATH,
            type=str,
        )
        isf_group = parser.add_mutually_exclusive_group()
        isf_group.add_argument(
            "--offline",
            help="Do not search online for additional JSON files",
            default=False,
            action="store_true",
        )
        isf_group.add_argument(
            "-u",
            "--remote-isf-url",
            metavar="URL",
            help="Search online for ISF json files",
            default=constants.REMOTE_ISF_URL,
            type=str,
        )
        parser.add_argument(
            "--filters",
            help="List of filters to apply to the output (in the form of [+-]columname,pattern[!])",
            default=[],
            action="append",
        )
        parser.add_argument(
            "--hide-columns",
            help="Case-insensitive space separated list of prefixes to determine which columns to hide in the output if provided",
            default=None,
            action="extend",
            nargs="*",
            type=str,
        )

        parser.set_defaults(**default_config)

        # We have to filter out help, otherwise parse_known_args will trigger the help message before having
        # processed the plugin choice or had the plugin subparser added.
        known_args = [arg for arg in sys.argv if arg != "--help" and arg != "-h"]
        partial_args, _ = parser.parse_known_args(known_args)

        ### Start up logging
        if partial_args.log:
            file_logger = logging.FileHandler(partial_args.log)
            file_logger.setLevel(1)
            file_formatter = logging.Formatter(
                datefmt="%y-%m-%d %H:%M:%S",
                fmt="%(asctime)s %(name)-12s %(levelname)-8s %(message)s",
            )
            file_logger.setFormatter(file_formatter)
            rootlog.addHandler(file_logger)
            vollog.info("Logging started")

        self.order_extra_verbose_levels()
        if partial_args.verbosity < 3:
            if partial_args.verbosity < 1:
                sys.tracebacklimit = None
            console.setLevel(logging.WARNING - (partial_args.verbosity * 10))
        else:
            console.setLevel(logging.DEBUG - (partial_args.verbosity - 2))

        for level, msg in delayed_logs:
            vollog.log(level, msg)

        ### Alter constants if necessary
        if partial_args.plugin_dirs:
            volatility3.plugins.__path__ = [
                os.path.abspath(p) for p in partial_args.plugin_dirs.split(";")
            ] + constants.PLUGINS_PATH

        if partial_args.symbol_dirs:
            volatility3.symbols.__path__ = [
                os.path.abspath(p) for p in partial_args.symbol_dirs.split(";")
            ] + constants.SYMBOL_BASEPATHS

        if partial_args.cache_path:
            constants.CACHE_PATH = partial_args.cache_path

        vollog.info(f"Volatility plugins path: {volatility3.plugins.__path__}")
        vollog.info(f"Volatility symbols path: {volatility3.symbols.__path__}")

        # Set the PARALLELISM
        if partial_args.parallelism == "processes":
            constants.PARALLELISM = constants.Parallelism.Multiprocessing
        elif partial_args.parallelism == "threads":
            constants.PARALLELISM = constants.Parallelism.Threading
        else:
            constants.PARALLELISM = constants.Parallelism.Off

        if partial_args.clear_cache:
            framework.clear_cache()

        if partial_args.offline:
            constants.OFFLINE = partial_args.offline
        elif partial_args.remote_isf_url:
            constants.REMOTE_ISF_URL = partial_args.remote_isf_url

        # Do the initialization
        ctx = contexts.Context()  # Construct a blank context
        failures = framework.import_files(
            volatility3.plugins, True
        )  # Will not log as console's default level is WARNING
        if failures:
            parser.epilog = (
                "The following plugins could not be loaded (use -vv to see why): "
                + ", ".join(sorted(failures))
            )
            vollog.info(parser.epilog)
        automagics = automagic.available(ctx)

        plugin_list = framework.list_plugins()

        # Discover renderers after plugin directories are loaded
        # This allows custom renderers to be found in plugin directories
        renderers = dict(
            [
                (x.name.lower(), x)
                for x in framework.class_subclasses(text_renderer.CLIRenderer)
            ]
        )

        parser.add_argument(
            "-r",
            "--renderer",
            metavar="RENDERER",
            help=f"Determines how to render the output ({', '.join(list(renderers))})",
            default="quick",
            choices=list(renderers),
        )

        seen_automagics = set()
        chosen_configurables_list = {}
        for amagic in automagics:
            if amagic in seen_automagics:
                continue
            seen_automagics.add(amagic)
            if isinstance(amagic, interfaces.configuration.ConfigurableInterface):
                self.populate_requirements_argparse(parser, amagic.__class__)

        subparser = parser.add_subparsers(
            title="Plugins",
            dest="plugin",
            description=f"For plugin specific options, run '{self.CLI_NAME} <plugin> --help'",
            action=volargparse.HelpfulSubparserAction,
            metavar="PLUGIN",
        )
        for plugin in sorted(plugin_list):
            # First line of a plugin docstring will be the short description for -h.
            # Text after the first two consecutive new lines will be
            # the additional description (argparse epilog).
            short_help = additional_help = None
            if plugin_list[plugin].__doc__ is not None:
                doc_split = plugin_list[plugin].__doc__.split("\n\n", 1)
                short_help = doc_split[0].strip()
                if len(doc_split) > 1:
                    additional_help = doc_split[1].strip()

            plugin_parser = subparser.add_parser(
                plugin,
                help=short_help,
                description=short_help,
                epilog=additional_help,
            )
            self.populate_requirements_argparse(plugin_parser, plugin_list[plugin])

        ###
        # PASS TO UI
        ###
        # Hand the plugin requirements over to the CLI (us) and let it construct the config tree

        # Run the argparser
        if HAS_ARGCOMPLETE:
            # The autocompletion line must be after the partial_arg handling, so that it doesn't trip it
            # before all the plugins have been added
            argcomplete.autocomplete(parser)
        args = parser.parse_args()

        # Display banner - redirect to stderr if using structured output
        banner_output = sys.stdout
        if renderers[args.renderer].structured_output:
            banner_output = sys.stderr
        banner_output.write(f"Volatility 3 Framework {constants.PACKAGE_VERSION}\n")

        if args.plugin is None:
            parser.error(
                f"Please select a plugin to run (see '{self.CLI_NAME} --help' for options"
            )

        vollog.log(
            constants.LOGLEVEL_VVV, f"Cache directory used: {constants.CACHE_PATH}"
        )

        plugin = plugin_list[args.plugin]
        chosen_configurables_list[args.plugin] = plugin
        base_config_path = "plugins"
        plugin_config_path = interfaces.configuration.path_join(
            base_config_path, plugin.__name__
        )

        # Special case the -f argument because people use is so frequently
        # It has to go here so it can be overridden by single-location if it's defined
        # NOTE: This will *BREAK* if LayerStacker, or the automagic configuration system, changes at all
        ###
        if args.file:
            try:
                single_location = requirements.URIRequirement.location_from_file(
                    args.file
                )
                ctx.config["automagic.LayerStacker.single_location"] = single_location
            except ValueError as excp:
                parser.error(str(excp))

        # UI fills in the config, here we load it from the config file and do it before we process the CL parameters
        if args.config:
            with open(args.config) as f:
                json_val = json.load(f)
                ctx.config.splice(
                    plugin_config_path,
                    interfaces.configuration.HierarchicalDict(json_val),
                )

        # It should be up to the UI to determine which automagics to run, so this is before BACK TO THE FRAMEWORK
        automagics = automagic.choose_automagic(automagics, plugin)
        for amagic in automagics:
            chosen_configurables_list[amagic.__class__.__name__] = amagic

        if ctx.config.get("automagic.LayerStacker.stackers", None) is None:
            ctx.config["automagic.LayerStacker.stackers"] = stacker.choose_os_stackers(
                plugin
            )
        self.output_dir = args.output_dir
        if not os.path.exists(self.output_dir):
            parser.error(
                f"The output directory specified does not exist: {self.output_dir}"
            )

        self.populate_config(ctx, chosen_configurables_list, args, plugin_config_path)

        if args.extend:
            for extension in args.extend:
                if "=" not in extension:
                    raise ValueError(
                        "Invalid extension (extensions must be of the format \"conf.path.value='value'\")"
                    )
                address, value = (
                    extension[: extension.find("=")],
                    json.loads(extension[extension.find("=") + 1 :]),
                )
                ctx.config[address] = value

        ###
        # BACK TO THE FRAMEWORK
        ###
        constructed = None
        try:
            progress_callback = PrintedProgress()
            if args.quiet:
                progress_callback = MuteProgress()

            constructed = plugins.construct_plugin(
                ctx,
                automagics,
                plugin,
                base_config_path,
                progress_callback,
                self.file_handler_class_factory(),
            )

            if args.write_config:
                vollog.warning(
                    "Use of --write-config has been deprecated, replaced by --save-config <filename>"
                )
                args.save_config = "config.json"
            if args.save_config:
                vollog.debug("Writing out configuration data to {args.save_config}")
                if os.path.exists(os.path.abspath(args.save_config)):
                    parser.error(
                        f"Cannot write configuration: file {args.save_config} already exists"
                    )
                with open(args.save_config, "w") as f:
                    json.dump(
                        dict(constructed.build_configuration()),
                        f,
                        sort_keys=True,
                        indent=2,
                    )
                    f.write("\n")
        except exceptions.UnsatisfiedException as excp:
            self.process_unsatisfied_exceptions(excp)
            parser.exit(
                1,
                f"Unable to validate the plugin requirements: {[x for x in excp.unsatisfied]}\n",
            )

        try:
            # Construct and run the plugin
            if constructed:
                vollog.debug(
                    f"Successfully constructed {args.plugin} {constructed.version}"
                )
                grid = constructed.run()
                renderer = renderers[args.renderer]()
                renderer.filter = text_filter.CLIFilter(grid, args.filters)
                renderer.column_hide_list = args.hide_columns
                renderer.render(grid)
        except exceptions.VolatilityException as excp:
            self.process_exceptions(excp)

    @classmethod
    def location_from_file(cls, filename: str) -> str:
        """Returns the URL location from a file parameter (which may be a URL)

        Args:
            filename: The path to the file (either an absolute, relative, or URL path)

        Returns:
            The URL for the location of the file
        """
        vollog.debug(
            f"{__name__}.location_from_file has been deprecated and moved to requirements.URIRequirement.location_from_file"
        )
        return requirements.URIRequirement.location_from_file(filename)

    def load_system_defaults(
        self, filename: str
    ) -> Tuple[List[Tuple[int, str]], Dict[str, Any]]:
        """Modify the main configuration based on the default configuration override"""
        # Build the config path
        default_config_path = os.path.join(
            os.path.expanduser("~"), ".config", "volatility3", filename
        )
        if sys.platform == "win32":
            default_config_path = os.path.join(
                os.environ.get("APPDATA", os.path.expanduser("~")),
                "volatility3",
                filename,
            )

        delayed_logs = []

        # Process it if the files exist
        if os.path.exists(default_config_path):
            with open(default_config_path, "rb") as config_json:
                result = json.load(config_json)
            if not isinstance(result, dict):
                delayed_logs.append(
                    (
                        logging.INFO,
                        f"Default configuration file {default_config_path} does not contain a dictionary",
                    )
                )
            else:
                delayed_logs.append(
                    (
                        logging.INFO,
                        f"Loading default configuration options from {default_config_path}",
                    )
                )
                delayed_logs.append(
                    (
                        logging.DEBUG,
                        f"Loaded configuration: {json.dumps(result, indent=2, sort_keys=True)}",
                    )
                )
                return delayed_logs, result
        return delayed_logs, {}

    def process_exceptions(self, excp):
        """Provide useful feedback if an exception occurs during a run of a plugin."""
        # Ensure there's nothing in the cache
        sys.stdout.write("\n\n")
        sys.stdout.flush()
        sys.stderr.flush()

        # Log the full exception at a high level for easy access
        fulltrace = traceback.TracebackException.from_exception(excp).format(chain=True)
        vollog.debug("".join(fulltrace))

        file_a_bug_msg = f"Please re-run with -vvv and file a bug with the output at {constants.BUG_URL}"

        if isinstance(excp, exceptions.InvalidAddressException):
            general = "Volatility was unable to read a requested page:"
            if isinstance(excp, exceptions.SwappedInvalidAddressException):
                detail = f"Swap error {hex(excp.invalid_address)} in layer {excp.layer_name} ({excp})"
                caused_by = [
                    "No suitable swap file having been provided (locate and provide the correct swap file)",
                    "An intentionally invalid page (operating system protection)",
                ]
            elif isinstance(excp, exceptions.PagedInvalidAddressException):
                detail = f"Page error {hex(excp.invalid_address)} in layer {excp.layer_name} ({excp})"
                caused_by = [
                    "Memory smear during acquisition (try re-acquiring if possible)",
                    "An intentionally invalid page lookup (operating system protection)",
                    "A bug in the plugin/volatility3 (re-run with -vvv and file a bug)",
                ]
            else:
                detail = (
                    f"{hex(excp.invalid_address)} in layer {excp.layer_name} ({excp})"
                )
                caused_by = [
                    "The base memory file being incomplete (try re-acquiring if possible)",
                    "Memory smear during acquisition (try re-acquiring if possible)",
                    "An intentionally invalid page lookup (operating system protection)",
                    "A bug in the plugin/volatility3 (re-run with -vvv and file a bug)",
                ]
        elif isinstance(excp, exceptions.SymbolError):
            general = "Volatility experienced a symbol-related issue:"
            detail = f"{excp.table_name}{constants.BANG}{excp.symbol_name}: {excp}"
            caused_by = [
                "An invalid symbol table",
                "A plugin requesting a bad symbol",
                "A plugin requesting a symbol from the wrong table",
            ]
        elif isinstance(excp, exceptions.SymbolSpaceError):
            general = "Volatility experienced an issue related to a symbol table:"
            detail = f"{excp}"
            caused_by = [
                "An invalid symbol table",
                "A plugin requesting a bad symbol",
                "A plugin requesting a symbol from the wrong table",
            ]
        elif isinstance(excp, exceptions.LayerException):
            general = f"Volatility experienced a layer-related issue: {excp.layer_name}"
            detail = f"{excp}"
            caused_by = [f"A faulty layer implementation. {file_a_bug_msg}"]
        elif isinstance(excp, exceptions.MissingModuleException):
            general = f"Volatility could not import a necessary module: {excp.module}"
            detail = f"{excp}"
            caused_by = [
                "A required python module is not installed (install the module and re-run)"
            ]
        elif isinstance(excp, exceptions.RenderException):
            general = "Volatility experienced an issue when rendering the output:"
            detail = f"{excp}"
            caused_by = ["An invalid renderer option, such as no visible columns"]
        elif isinstance(excp, exceptions.VersionMismatchException):
            general = "A version mismatch was detected between two components:"
            detail = f"{excp}"
            caused_by = [
                excp.failure_reason or "An outdated API caller, such as a method.",
                file_a_bug_msg,
            ]
        else:
            general = "Volatility encountered an unexpected situation."
            detail = ""
            caused_by = [file_a_bug_msg]

        # Code that actually renders the exception
        output = sys.stderr
        output.write(f"{general}\n")
        output.write(f"{detail}\n\n")
        for cause in caused_by:
            output.write(f"	* {cause}\n")
        output.write("\nNo further results will be produced\n")
        sys.exit(1)

    def process_unsatisfied_exceptions(self, excp):
        """Provide useful feedback if an exception occurs during requirement fulfillment."""
        # Add a blank newline
        print("")
        translation_failed = False
        symbols_failed = False
        for config_path in excp.unsatisfied:
            translation_failed = translation_failed or isinstance(
                excp.unsatisfied[config_path],
                configuration.requirements.TranslationLayerRequirement,
            )
            symbols_failed = symbols_failed or isinstance(
                excp.unsatisfied[config_path],
                configuration.requirements.SymbolTableRequirement,
            )

            print(
                f"Unsatisfied requirement {config_path}: {excp.unsatisfied[config_path].description}"
            )

        if translation_failed:
            print(
                "\nA translation layer requirement was not fulfilled.  Please verify that:\n"
                "\tA file was provided to create this layer (by -f, --single-location or by config)\n"
                "\tThe file exists and is readable\n"
                "\tThe file is a valid memory image and was acquired cleanly"
            )
        if symbols_failed:
            print(
                "\nA symbol table requirement was not fulfilled.  Please verify that:\n"
                "\tThe associated translation layer requirement was fulfilled\n"
                "\tYou have the correct symbol file for the requirement\n"
                "\tThe symbol file is under the correct directory or zip file\n"
                "\tThe symbol file is named appropriately or contains the correct banner\n"
            )

    def populate_config(
        self,
        context: interfaces.context.ContextInterface,
        configurables_list: Dict[
            str, Type[interfaces.configuration.ConfigurableInterface]
        ],
        args: argparse.Namespace,
        plugin_config_path: str,
    ) -> None:
        """Populate the context config based on the returned args.

        We have already determined these elements must be descended from ConfigurableInterface

        Args:
            context: The volatility3 context to operate on
            configurables_list: A dictionary of configurable items that can be configured on the plugin
            args: An object containing the arguments necessary
            plugin_config_path: The path within the context's config containing the plugin's configuration
        """
        vargs = vars(args)
        for configurable in configurables_list:
            for requirement in configurables_list[configurable].get_requirements():
                value = vargs.get(requirement.name, None)
                if value is not None:
                    if isinstance(requirement, requirements.URIRequirement):
                        if isinstance(value, str):
                            scheme = parse.urlparse(value).scheme
                            if not scheme or len(scheme) <= 1:
                                if not os.path.exists(value):
                                    raise FileNotFoundError(
                                        f"Non-existent file {value} passed to URIRequirement"
                                    )
                                value = f"file://{request.pathname2url(os.path.abspath(value))}"
                    if isinstance(requirement, requirements.ListRequirement):
                        if not isinstance(value, list):
                            raise TypeError(
                                f"Configuration for ListRequirement was not a list: {requirement.name}"
                            )
                        value = [requirement.element_type(x) for x in value]
                    if not inspect.isclass(configurables_list[configurable]):
                        config_path = configurables_list[configurable].config_path
                    else:
                        # We must be the plugin, so name it appropriately:
                        config_path = plugin_config_path
                    extended_path = interfaces.configuration.path_join(
                        config_path, requirement.name
                    )
                    context.config[extended_path] = value

    def order_extra_verbose_levels(self):
        for level, level_value in enumerate(
            [
                constants.LOGLEVEL_V,
                constants.LOGLEVEL_VV,
                constants.LOGLEVEL_VVV,
                constants.LOGLEVEL_VVVV,
            ]
        ):
            logging.addLevelName(level_value, f"DETAIL {level + 1}")

    def file_handler_class_factory(self, direct=True):
        output_dir = self.output_dir

        class CLIFileHandler(interfaces.plugins.FileHandlerInterface):
            def _get_final_filename(self):
                """Gets the final filename"""
                if output_dir is None:
                    raise TypeError("Output directory is not a string")

                os.makedirs(output_dir, exist_ok=True)

                output_filename = os.path.join(output_dir, self.preferred_filename)
                filename, extension = os.path.splitext(output_filename)

                counter = 1
                while os.path.exists(output_filename):
                    output_filename = f"{filename}-{counter}{extension}"
                    counter += 1

                return output_filename

        class CLIMemFileHandler(io.BytesIO, CLIFileHandler):
            def __init__(self, filename: str):
                io.BytesIO.__init__(self)
                CLIFileHandler.__init__(self, filename)

            def close(self):
                # Don't overcommit
                if self.closed:
                    return None

                self.seek(0)

                output_filename = self._get_final_filename()

                with open(output_filename, "wb") as current_file:
                    current_file.write(self.read())
                    self._committed = True
                    vollog.log(
                        logging.INFO, f"Saved stored plugin file: {output_filename}"
                    )

                super().close()

        class CLIDirectFileHandler(CLIFileHandler):
            def __init__(self, filename: str):
                fd, self._name = tempfile.mkstemp(
                    suffix=".vol3", prefix="tmp_", dir=output_dir
                )
                self._file = open(fd, mode="w+b")
                CLIFileHandler.__init__(self, filename)
                for item in dir(self._file):
                    if not item.startswith("_") and item not in (
                        "closed",
                        "close",
                        "mode",
                        "name",
                    ):
                        setattr(self, item, getattr(self._file, item))

            def __getattr__(self, item):
                return getattr(self._file, item)

            @property
            def closed(self):
                return self._file.closed

            @property
            def mode(self):
                return self._file.mode

            @property
            def name(self):
                return self._file.name

            def close(self):
                """Closes and commits the file (by moving the temporary file to the correct name"""
                # Don't overcommit
                if self._file.closed:
                    return None

                output_filename = self._get_final_filename()

                # Update the filename, which may have changed if a file with
                # the same name already existed. This needs to be done before
                # closing the file, otherwise FileHandlerInterface will raise
                # an exception. Also, the preferred_filename setter only allows
                #  a specific set of characters, where '/' is not in that list
                self.preferred_filename = os.path.basename(output_filename)

                self._file.close()
                os.rename(self._name, output_filename)

        if direct:
            return CLIDirectFileHandler
        else:
            return CLIMemFileHandler

    def populate_requirements_argparse(
        self,
        parser: Union[argparse.ArgumentParser, argparse._ArgumentGroup],
        configurable: Type[interfaces.configuration.ConfigurableInterface],
    ):
        """Adds the plugin's simple requirements to the provided parser.

        Args:
            parser: The parser to add the plugin's (simple) requirements to
            configurable: The plugin object to pull the requirements from
        """
        if not issubclass(configurable, interfaces.configuration.ConfigurableInterface):
            raise TypeError(
                f"Expected ConfigurableInterface type, not: {type(configurable)}"
            )

        # Construct an argparse group

        for requirement in configurable.get_requirements():
            additional: Dict[str, Any] = {}
            if not isinstance(
                requirement, interfaces.configuration.RequirementInterface
            ):
                raise TypeError(
                    f"Plugin contains requirements that are not RequirementInterfaces: {configurable.__name__}"
                )
            if isinstance(requirement, interfaces.configuration.SimpleTypeRequirement):
                additional["type"] = requirement.instance_type
                if isinstance(requirement, requirements.IntRequirement):
                    additional["type"] = lambda x: int(x, 0)
                if isinstance(requirement, requirements.BooleanRequirement):
                    additional["action"] = "store_true"
                    if "type" in additional:
                        del additional["type"]
            elif isinstance(
                requirement,
                volatility3.framework.configuration.requirements.ListRequirement,
            ):
                # Allow a list of integers, specified with the convenient 0x hexadecimal format
                if requirement.element_type is int:
                    additional["type"] = lambda x: int(x, 0)
                else:
                    additional["type"] = requirement.element_type
                nargs = "*" if requirement.optional else "+"
                additional["nargs"] = nargs
            elif isinstance(
                requirement,
                volatility3.framework.configuration.requirements.ChoiceRequirement,
            ):
                additional["type"] = str
                additional["choices"] = requirement.choices
            else:
                continue
            parser.add_argument(
                "--" + requirement.name.replace("_", "-"),
                help=requirement.description,
                default=requirement.default,
                dest=requirement.name,
                required=not requirement.optional,
                **additional,
            )


def main():
    """A convenience function for constructing and running the
    :class:`CommandLine`'s run method."""
    CommandLine().run()



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

from typing import Any, List, Optional, Tuple, Union

from volatility3.cli.volshell import generic
from volatility3.framework import constants, interfaces
from volatility3.framework.configuration import requirements
from volatility3.plugins.mac import pslist


class Volshell(generic.Volshell):
    """Shell environment to directly interact with a mac memory image."""

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel", description="Darwin kernel module"
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.IntRequirement(
                name="pid", description="Process ID", optional=True
            ),
            requirements.VersionRequirement(
                name="generic_volshell",
                component=generic.Volshell,
                version=(1, 0, 0),
            ),
        ] + super().get_requirements()

    def change_task(self, pid=None):
        """Change the current process and layer, based on a process ID"""
        tasks = self.list_tasks()
        for task in tasks:
            if task.p_pid == pid:
                process_layer = task.add_process_layer()
                if process_layer is not None:
                    self.change_layer(process_layer)
                    return None
                print(f"Layer for task ID {pid} could not be constructed")
                return None
        print(f"No task with task ID {pid} found")

    def list_tasks(self, method=None):
        """Returns a list of task objects from the primary layer"""
        # We always use the main kernel memory and associated symbols
        return list(
            pslist.PsList.get_list_tasks(method)(self.context, self.current_kernel_name)
        )

    def construct_locals(self) -> List[Tuple[List[str], Any]]:
        result = super().construct_locals()
        result += [
            (["ct", "change_task", "cp"], self.change_task),
            (["lt", "list_tasks", "ps"], self.list_tasks),
            (["symbols"], self.context.symbol_space[self.current_symbol_table]),
        ]
        if self.config.get("pid", None) is not None:
            self.change_task(self.config["pid"])
        return result

    def display_type(
        self,
        object: Union[
            str, interfaces.objects.ObjectInterface, interfaces.objects.Template
        ],
        offset: Optional[int] = None,
    ):
        """Display Type describes the members of a particular object in alphabetical order"""
        if isinstance(object, str):
            if constants.BANG not in object:
                object = self.current_symbol_table + constants.BANG + object
        return super().display_type(object, offset)

    def display_symbols(self, symbol_table: Optional[str] = None):
        """Prints an alphabetical list of symbols for a symbol table"""
        if symbol_table is None:
            symbol_table = self.current_symbol_table
        return super().display_symbols(symbol_table)

    @property
    def current_layer(self):
        if self.__current_layer is None:
            self.__current_layer = self.kernel.layer_name
        return self.__current_layer



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""Defines the symbols architecture.

This is the namespace for all volatility symbols, and determines the
path for loading symbol ISF files
"""

from volatility3.framework import constants

__path__ = constants.SYMBOL_BASEPATHS



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""A list of potential exceptions that volatility can throw.

These include exceptions that can be thrown on errors by the symbol
space or symbol tables, and by layers when an address is invalid.  The
:class:`PagedInvalidAddressException` contains information about the
size of the invalid page.
"""

from typing import Callable, Dict, Optional, Tuple

from volatility3.framework import interfaces


class VolatilityException(Exception):
    """Class to allow filtering of all VolatilityExceptions."""


class PluginVersionException(VolatilityException):
    """Class to allow determining that a required plugin has an invalid
    version."""


class PluginRequirementException(VolatilityException):
    """Class to allow plugins to indicate that a requirement has not been
    fulfilled."""


class SymbolError(VolatilityException):
    """Thrown when a symbol lookup has failed."""

    def __init__(
        self, symbol_name: Optional[str], table_name: Optional[str], *args
    ) -> None:
        super().__init__(*args)
        self.symbol_name = symbol_name
        self.table_name = table_name


class LayerException(VolatilityException):
    """Thrown when an error occurs dealing with memory and layers."""

    def __init__(self, layer_name: str, *args) -> None:
        super().__init__(*args)
        self.layer_name = layer_name


class InvalidAddressException(LayerException):
    """Thrown when an address is not valid in the layer it was requested."""

    def __init__(self, layer_name: str, invalid_address: int, *args) -> None:
        super().__init__(layer_name, *args)
        self.invalid_address = invalid_address


class PagedInvalidAddressException(InvalidAddressException):
    """Thrown when an address is not valid in the paged space in which it was
    request.  This is a subclass of InvalidAddressException and is only
    thrown from a paged layer.  In most circumstances :class:`InvalidAddressException`
    is the correct exception to throw, since this will catch all invalid
    mappings (including paged ones).

    Includes the invalid address and the number of bits of the address
    that are invalid
    """

    def __init__(
        self,
        layer_name: str,
        invalid_address: int,
        invalid_bits: int,
        entry: int,
        *args,
    ) -> None:
        super().__init__(layer_name, invalid_address, *args)
        self.invalid_bits = invalid_bits
        self.entry = entry


class SwappedInvalidAddressException(PagedInvalidAddressException):
    """Thrown when an address is not valid in the paged layer in which it was
    requested, but expected to be in an associated swap layer.

    Includes the swap lookup, as well as the invalid address and the bits of
    the lookup that were invalid.
    """

    def __init__(
        self,
        layer_name: str,
        invalid_address: int,
        invalid_bits: int,
        entry: int,
        swap_offset: int,
        *args,
    ) -> None:
        super().__init__(layer_name, invalid_address, invalid_bits, entry, *args)
        self.swap_offset = swap_offset


class SymbolSpaceError(VolatilityException):
    """Thrown when an error occurs dealing with Symbolspaces and SymbolTables."""


class UnsatisfiedException(VolatilityException):
    def __init__(
        self, unsatisfied: Dict[str, interfaces.configuration.RequirementInterface]
    ) -> None:
        super().__init__()
        self.unsatisfied = unsatisfied


class MissingModuleException(VolatilityException):
    def __init__(self, module: str, *args) -> None:
        super().__init__(*args)
        self.module = module


class OfflineException(VolatilityException):
    """Throw when a remote resource is requested but Volatility is in offline mode"""

    def __init__(self, url: str, *args) -> None:
        super().__init__(*args)
        self._url = url

    def __str__(self):
        return f"Volatility 3 is offline: unable to access {self._url}"


class RenderException(VolatilityException):
    """Thrown if there is an error during rendering"""


class LinuxPageCacheException(VolatilityException):
    """Thrown if there is an error during Linux Page Cache processing"""


class VersionMismatchException(VolatilityException):
    """Thrown if a version mismatch has been encountered between two components."""

    def __init__(
        self,
        source_component: Callable,
        target_component: interfaces.configuration.VersionableInterface,
        target_version: Tuple[int, int, int],
        failure_reason: str = None,
        *args,
    ):
        """
        Args:
            source_component: The component that required the target component
            target_component: The component that is required. Must inherit from interfaces.configuration.VersionableInterface
            target_version: The version of the target component that was required, and ultimately was not satisfied
            failure_reason: A detailed failure reason to enhance debugging and bug tracking
        """
        super().__init__(*args)
        self.source_component = source_component
        self.target_component = target_component
        self.target_version = target_version
        self.failure_reason = failure_reason

    def __str__(self):
        return f"{self.source_component.__module__ + '.' + self.source_component.__qualname__}: Version {self.target_version} dependency on {self.target_component.__module__ + '.' + self.target_component.__name__} {self.target_component.version} unmet."
